"""
Tests for base storage implementation, to ensure it must always be overridden
"""
import pytest

from autossl.storage import base


@pytest.fixture(scope="function")
def base_storage():
    yield base.Storage(tracking_record_id=None)


def test_save_data(base_storage):
    with pytest.raises(NotImplementedError):
        base_storage.save_data(name='my_data_name', data_type=None, content=None, local_path=None)


def test_retrieve_data(base_storage):
    with pytest.raises(NotImplementedError):
        base_storage.retrieve_data(name='my_file_name', data_type=None)



from enum import Enum
import getpass
import os

from . import util


class CredentialType(Enum):
    """list of credentials types supported"""
    UserPassword = 'user_password'
    ApiKeyAndId = 'api_key_and_api_id'


def get_credentials(name, global_config, credentials, extra_parameters=None):
    """Get structured form of specified credential based on its type and ready to be passed to any api

    :param name: name of the credential
    :type name: str
    :param global_config: credential global configuration
    :type global_config: dict
    :param credentials: structured credentials  dict
    :type credentials: dict
    :param extra_parameters: extra parameters to add to current credential
    :type extra_parameters: dict
    :return: structured credentials
    :rtype: dict
    """
    result = credentials.get(name, {})
    if result.get('type') is None:
        # retrieve credentials global configuration
        credentials_config = global_config[name]

        # retrieve credentials based on credentials type
        # credential type should be valid as already verified as part of ssl blueprint validation
        credential_type = CredentialType[credentials_config['type']]

        # ability to customize fields separator in command line from global credential section, default to ':'
        separator = credentials_config.get('separator')

        # no need for an 'else' as every credential type part of the enum is expected to be implemented here
        if credential_type == CredentialType.UserPassword:
            result = get_user_password(name=name, credentials=credentials, separator=separator)
        elif credential_type == CredentialType.ApiKeyAndId:
            result = get_api_key_and_id(name=name, credentials=credentials, separator=separator)

        # update custom parameters
        if extra_parameters:
            result.update(extra_parameters)

    return result


def _get_credential(name, credentials, credential_type, separator, parameters, sensitive_params=None):
    # initialize result with correct structure this credential type
    result = {
        'type': credential_type,
    }
    for parameter in parameters:
        result[parameter] = None

    # check first if this credential is already in input credentials dict,
    # this can happen in 2 cases:
    # 1. if provided in command line input
    # 2. if it has been already used once
    if credentials and name in credentials:
        result.update(credentials[name])

    # credential is not complete, let try different ways to get missing information
    missing_parameters = [param_name for param_name in result if result[param_name] is None]
    if any(missing_parameters):
        # first, if we have a raw content to parse
        raw_content = result.get('raw_content')
        if raw_content is not None:
            # try to split as much as we expect parameters, using specified separator or ':' by default
            credentials_input = raw_content.split(separator or ':', len(parameters) - 1)
            for index, parameter in enumerate(parameters):
                if len(credentials_input) >= index + 1:
                    result[parameter] = credentials_input[index]
            # only keep formatted data
            del result['raw_content']

        # finally if we still miss some data
        for parameter in parameters:
            if result[parameter] is None:
                # check environment variables first
                # format for env variable name is "CREDENTIAL_NAME + _ + PARAMETER_NAME"
                result[parameter] = os.getenv('{}_{}'.format(name, parameter).upper())

            # and finally, if still missing, ask interactively to the user
            if result[parameter] is None:
                prompt_str = '{} {}:'.format(name.title(), parameter)
                if parameter in sensitive_params or []:
                    result[parameter] = getpass.getpass(prompt_str)
                else:
                    result[parameter] = raw_input(prompt_str) if util.PY2 else input(prompt_str)  # noqa

    return result


def get_user_password(name, credentials=None, separator=None):
    return _get_credential(
        name=name,
        credentials=credentials,
        credential_type=CredentialType.UserPassword,
        separator=separator,
        parameters=['username', 'password'],
        sensitive_params=['password']
    )


def get_api_key_and_id(name, credentials=None, separator=None):
    return _get_credential(
        name=name,
        credentials=credentials,
        separator=separator,
        credential_type=CredentialType.ApiKeyAndId,
        parameters=['api_key', 'api_id'],
        sensitive_params=['api_key', 'api_id']
    )



from enum import Enum
import logging

logger = logging.getLogger(__name__)


class TrackingType(Enum):
    """list of tracking types supported."""
    # End-to-end flow of certificate renewal and deployment on servers
    Renewal = 'renewal'
    # Simply deploy existing valid certificate on new or outdated servers
    Synchronize = 'synchronize'


class Tracking(object):

    def __init__(self, ssl_blueprint_path, **kwargs):
        r"""Api to tracking server for specified input ssl blueprint

        :param ssl_blueprint_path: local path to ssl blueprint
        :type ssl_blueprint_path: pathlib.Path
        :param \**kwargs: generic key/value parameters
        :type \**kwargs: dict
        """
        self.ssl_blueprint_path = ssl_blueprint_path

    def create(self, tracking_type, servers=None):
        """Create a tracking record with details of current SSL blueprint

        :param tracking_type: Type of tracking. Can be used to customized tracking record content.
        :type tracking_type: TrackingType
        :param servers: List of servers in scope of the action. All servers from config if None specified here.
        :type servers: list
        :return: Identifier for the created record
        :rtype: str
        """
        logger.debug("Nothing to do for 'create' tracking default implementation.")

    def save_data(self, name, data_type, local_path=None, content=None, **kwargs):
        r"""Save input data in tracking system

        :param name: name of the file to attach to the tracking record
        :type name: str
        :param data_type: type of data to save
        :type data_type: ssl.DataType
        :param local_path: local path to file to attach to the tracking record
        :type local_path: pathlib.Path
        :param content: content of the file to attach to the tracking record
        :type content: bytes
        :param \**kwargs: generic key/value parameters
        :type kwargs: dict
        """
        logger.debug("Nothing to do for 'save_data' tracking default implementation.")

    def update(self, message):
        """Update tracking record

        :param message: text to add to tracking record
        :type message: str
        """
        logger.debug("Nothing to do for 'update' tracking default implementation.")

    def refresh(self, record_id):
        """Update current tracking instance with last changes from tracking record on server side

        :param record_id: identifier of the record to refresh
        """
        logger.debug("Nothing to do for 'refresh' tracking default implementation.")

    def retrieve_data(self, name=None, data_type=None, **kwargs):
        r"""Retrieve specified data from tracking system

        :param name: Name of file/data to retrieve
        :type name: str
        :param data_type: type of data to retrieve
        :type data_type: ssl.DataType
        :param \**kwargs: generic key/value parameters
        :type kwargs: dict
        :return: file content
        :rtype: bytes
        """
        logger.debug("Nothing to do for 'retrieve_data' tracking default implementation.")

    def close_for_failure(self, message):
        """Specify action is completed with a failed status

        :param message: custom message
        :type message: str
        """
        logger.debug("Nothing to do for 'close_for_failure' tracking default implementation.")

    def close_for_success(self, message):
        """Specify action is completed with a success status

        :param message: custom message
        :type message: str
        """
        logger.debug("Nothing to do for 'close_for_success' tracking default implementation.")



import logging
import os
import shutil

from .. import exception, util
from . import base

logger = logging.getLogger(__name__)


class LocalFileStorage(base.Storage):

    def __init__(self, path, tracking_record_id=None, **kwargs):
        super(LocalFileStorage, self).__init__(tracking_record_id=tracking_record_id, **kwargs)
        self.path = util.Path(os.path.expandvars(path))
        if not self.path.is_dir():
            raise IOError("Invalid folder path specified: '%s'" % self.path)

    def save_data(self, name, content=None, local_path=None, **kwargs):
        output_file_path = self.path / name

        if content is not None:
            if local_path:
                logger.warning("local_path path '{}' ignored as content also specified.".format(local_path))
            output_file_path.write_bytes(content)
        else:
            shutil.copy(str(local_path), str(output_file_path))

    def retrieve_data(self, name, **kwargs):
        file_path = self.path / name
        if not file_path.exists():
            raise exception.NotFound("Path %s does not exists." % file_path)
        return file_path.read_bytes()



"""
Tests for local storage implementation
"""
import os
import pytest

from autossl import exception
from autossl.storage import local

from tests import util as tests_util


@pytest.fixture(scope="function")
def local_storage(tmp_path):
    yield local.LocalFileStorage(path=str(tmp_path))


def test_init_error():
    with pytest.raises(IOError):
        local.LocalFileStorage(path=os.path.join(os.getcwd(), 'my', 'missing', 'path'))


def test_retrieve_data_error(local_storage):
    with pytest.raises(exception.NotFound):
        local_storage.retrieve_data(name='my_file_name')


def test_save_retrieve_data_from_content(local_storage):
    file_name = tests_util.get_random_ascii_string()
    file_content = tests_util.get_random_text()

    # save data in main repo from content
    local_storage.save_data(name=file_name,
                            content=file_content,
                            # local_path here will be ignored as content is already provided
                            local_path='not_existing_path')
    assert local_storage.retrieve_data(name=file_name) == file_content


def test_save_retrieve_data_from_local_file(tmp_path, local_storage):
    file_name = tests_util.get_random_ascii_string()
    file_content = tests_util.get_random_text()

    # save data with content from local file
    # first create a temporary file on disk that will be used as source of the data
    source_file_path = tmp_path / file_name
    source_file_path.write_bytes(file_content)

    # save data
    file_name2 = tests_util.get_random_ascii_string()
    local_storage.save_data(name=file_name2, local_path=source_file_path)

    assert local_storage.retrieve_data(name=file_name2) == file_content

    local_storage.save_data(name=file_name2, local_path=source_file_path)
    assert local_storage.retrieve_data(name=file_name2) == file_content



import collections
import shutil
import socket
import tempfile

import pytest

from autossl import exception, ssl, util
from autossl.server import local

from tests import util as tests_util


CertificateKeyPair = collections.namedtuple('CertificateKeyPair', 'key crt')


@pytest.fixture(scope="function")
def crt_dir(tmp_path):
    return tmp_path


@pytest.fixture(scope="function")
def local_server(tmp_path, crt_dir):
    # return local server instance
    yield local.LocalServer(crt_name='autossl.example.com', path=crt_dir, acme_dir=tmp_path)


@pytest.fixture(scope="module")
def ca_keypair_path():
    key, crt = tests_util.create_ca_certificate(ca_name='Autossl')
    ca_temp_dir = util.Path(tempfile.mkdtemp())
    ca_crt_path = ca_temp_dir / 'local_ca.crt'
    ca_key_path = ca_temp_dir / 'local_ca.key'

    ca_crt_path.write_bytes(crt)
    ca_key_path.write_bytes(key)

    yield CertificateKeyPair(ca_key_path, ca_crt_path)

    # cleanup temp folders
    shutil.rmtree(str(ca_temp_dir), ignore_errors=True)


@pytest.fixture(scope="function")
def keypair(tmp_path, ca_keypair_path):
    common_name = 'tst.autossl.example.com'
    crt_name = tests_util.get_random_ascii_string()
    key_path = tmp_path.joinpath(crt_name + '.key')
    key_content, csr_path = ssl.generate_csr(name=crt_name, common_name=common_name, output_path=tmp_path)
    crt_path = tmp_path.joinpath(crt_name + '.crt')
    crt_content = tests_util.create_signed_certificate(
        csr_path=csr_path,
        ca_crt_path=ca_keypair_path.crt,
        ca_key_path=ca_keypair_path.key,
        certificate_validity_days=10,
    )

    crt_path.write_bytes(crt_content)
    key_path.write_bytes(key_content)

    yield CertificateKeyPair(key_path, crt_path)


def test_get_description(local_server):
    assert local_server.get_description() == '[LocalServer - {}:{}]'.format(socket.gethostname(), local_server.path)


def test_deploy_cert(local_server, keypair):
    # deploy key + cert + ca cert
    local_server.deploy_cert(key=keypair.key, cert=keypair.crt)

    # ensure files have been created on local server and with proper content
    for source_path in [keypair.key, keypair.crt]:
        server_path = local_server.path.joinpath(local_server.crt_name + source_path.suffix)
        assert server_path.exists()
        assert source_path.read_bytes() == server_path.read_bytes()


@pytest.mark.parametrize('crt_dir', ['/etc/autossl/dummydir12345678'])
def test_deploy_cert_error(local_server, keypair):
    with pytest.raises(exception.DeployCertificateError) as ex_info:
        local_server.deploy_cert(key=keypair.key, cert=keypair.crt)
    assert 'No such file or directory' in str(ex_info.value)


def test_get_certificate_information(local_server, keypair):
    with pytest.raises(exception.CertificateNotFound):
        local_server.get_certificate_information()

    local_server.deploy_cert(key=keypair.key, cert=keypair.crt)

    assert local_server.get_certificate_information() == ssl.SslCertificate().init_from_x509(x509_path=keypair.crt)


def test_create_acme_challenge(local_server):
    # prepare data
    token = tests_util.get_random_ascii_string()
    key_authorization = tests_util.get_random_ascii_string()
    wellknown_path = local_server.acme_dir / token

    # check challenge does not exist yet
    assert not wellknown_path.exists()

    # deploy challenge
    local_server.create_acme_challenge(token=token, key_authorization=key_authorization)

    # ensure challenge exist and with proper value
    assert wellknown_path.exists()
    assert wellknown_path.read_text() == key_authorization


def test_delete_acme_challenge(local_server):
    # prepare data
    token = tests_util.get_random_ascii_string()
    key_authorization = tests_util.get_random_ascii_string()
    wellknown_path = local_server.acme_dir / token

    # deploy challenge
    local_server.create_acme_challenge(token=token, key_authorization=key_authorization)

    # check challenge exists
    assert wellknown_path.exists()

    # delete challenge
    local_server.delete_acme_challenge(token=token)

    # ensure challenge has been deleted
    assert not wellknown_path.exists()



import base64
import datetime
import json
import logging
import time

import requests

from .. import exception, ssl
from . import base

logger = logging.getLogger(__name__)


class IncapsulaSite(base.Server):

    BASE_URL = "https://my.incapsula.com:443"

    def __init__(self, api_key, api_id, site_id, crt_name, **kwargs):
        super(IncapsulaSite, self).__init__(crt_name=crt_name, **kwargs)
        self.api_key = api_key
        self.api_id = api_id
        self.site_id = site_id
        # The certificate must contain the full chain - root CA, intermediate CA, and the origin server certificates.
        # see https://docs.imperva.com/bundle/cloud-application-security/page/more/upload-ssl.htm
        self.deploy_full_chain = True

        # enable retry on :
        # - failed DNS lookups
        # - socket connections
        # - connection timeouts
        # but never to requests where data has made it to the server
        self.session = requests.Session()
        self.session.mount('https://', requests.adapters.HTTPAdapter(max_retries=5))

        # all queries must contain at least this parameters
        self.basic_params = {'api_key': self.api_key, 'api_id': self.api_id, 'site_id': self.site_id}
        self.site_status = self.session.post(
            url='{}/api/prov/v1/sites/status'.format(self.BASE_URL),
            data=self.basic_params,
        ).json()

    def get_description(self):
        return "[{} - {} ({})]".format(self.__class__.__name__, self.site_status['domain'], self.site_id)

    def deploy_cert(self, key, cert, **kwargs):
        # authentication data and site ID
        parameters = self.basic_params.copy()

        parameters.update({
            'certificate': base64.b64encode(cert.read_bytes()),
            'private_key': base64.b64encode(key.read_bytes()),
        })

        try:
            json_response = {}
            for _ in range(5):
                json_response = self.session.post(
                    url='{}/api/prov/v1/sites/customCertificate/upload'.format(self.BASE_URL),
                    data=parameters,
                ).json()
                # deployment is a success
                if json_response['res'] == 0:
                    break
                # specific behavior for Internal error code, that is returned quite often by Incapsula...
                # => just retry, it generally works fine the 2nd or 3rd time
                elif json_response['res'] == 3015:
                    time.sleep(1)
                    continue
            # no success, just return exception with last error message
            else:
                raise exception.DeployCertificateError('Unable to deploy new certificate on {}: {}'.format(
                    self.get_description(), json.dumps(json_response, indent=4)))

            # Upload successful
            logger.info("Certificate/Key %s updated successfully on %s.", self.crt_name, self.get_description())

        except requests.exceptions.RequestException as ex:
            raise exception.DeployCertificateError(
                msg='Unable to deploy new certificate on {}: {}'.format(self.get_description(), str(ex)),
                original_exception=ex,
            )

    def is_same(self, common_name=None, sans=None, exact_match=False):
        """Check if domain for targeted Incapsula site is part of specified domains

        :param common_name: Common name
        :type common_name: str
        :param sans: list of Subject Alternate Names
        :type sans: list
        :param exact_match: if True, certificate must exactly match input domains
            if False, input domain will also match wildcard certificate and additional domains in certificate will
            be ignored
        :type exact_match: bool
        :return: True if Incapsula site domain is covered by input domains
        """
        blueprint_domains = ssl.get_domains(common_name=common_name, sans=sans)

        # check if Incapsula domain is included in input domains
        for blueprint_domain in blueprint_domains:
            if ssl.is_domain_matching(domain_to_check=self.site_status['domain'],
                                      reference_domain=blueprint_domain,
                                      exact_match=exact_match):
                return True

        return False

    def get_certificate_information(self):
        """Retrieve certificate information from Incapsula site.

        :return: SSL certificate information
        :rtype: autossl.ssl.SslCertificate
        :raise autossl.exception.CertificateNotFound: if certificate does not exist yet on server
        """
        custom_certificate = self.site_status.get('ssl', {}).get('custom_certificate')

        # if invalid incapsula response or no custom_certificate deployed yet
        if self.site_status['res'] != 0 or custom_certificate is None or custom_certificate.get('active') is not True:
            raise exception.CertificateNotFound("No certificate found for site ID {}".format(self.get_description()))

        # Get expiration date (in milliseconds, since 1970) from site configuration
        expiration_date_ms = custom_certificate.get('expirationDate')
        if expiration_date_ms is None:
            raise RuntimeError(
                "Unable to get certificate expiration date (path: ssl.custom_certificate.expirationDate)"
                " for site ID {} in api response {}".format(self.get_description(),
                                                            json.dumps(self.site_status, indent=4))
            )

        return ssl.SslCertificate(
            common_name=self.site_status['domain'],
            sans=[],
            expiration=datetime.datetime.utcfromtimestamp(float(expiration_date_ms)/1000),
        )

    def create_acme_challenge(self, token, key_authorization):
        """Create token on server with specified value

        :param token: challenge key
        :param key_authorization: challenge value
        """
        logger.debug("No challenge to deploy for Incapsula that is just acting as a proxy to real server.")

    def delete_acme_challenge(self, token):
        """Delete challenge created on server

        :param token: challenge key to delete from server
        """
        logger.debug("No challenge to cleanup for Incapsula that is just acting as a proxy to real server.")



#!/usr/bin/env python
import logging
import os
import shutil

import git

from .. import exception, util
from . import base

logger = logging.getLogger(__name__)


def git_url_with_username_password(git_url, username, password):
    # This part handles git clone via http and not ssh. It is strongly advised to use
    # user/password combination when cloning in https. Most servers don't accept anonymous calls anymore
    if username and password:
        if 'http' in git_url:
            return git_url.replace('://', '://{}:{}@'.format(
                username, password
            ))
    else:
        return git_url


class GitStorage(base.Storage):

    def __init__(self, git_url, folder=None, tracking_record_id=None,
                 config_user_name=None, config_user_email=None, **kwargs):
        """Store data in git remote server

        :param git_url: full url of the repository to store data in
        :type git_url: str
        :param folder: path
        :type folder: pathlib.Path or str
        :param config_user_name: git config user name
        :type config_user_name: str
        :param config_user_email: git config user email
        :type config_user_email: str
        """
        super(GitStorage, self).__init__(tracking_record_id=tracking_record_id, **kwargs)
        self.folder = folder
        if folder is not None and not isinstance(folder, util.Path):
            self.folder = util.Path(folder)
        self.config_user_name = config_user_name
        self.config_user_email = config_user_email
        self.git_url = git_url_with_username_password(git_url,
                                                      kwargs.get('username'),
                                                      kwargs.get('password'))

    def save_data(self, name, content=None, local_path=None, **kwargs):
        with util.TempDir() as temp_folder:
            try:
                file_path, temp_file_path, git_repo = self.__clone(folder=temp_folder.path, name=name)

                # make sure parent folder and sub-folders exist
                parent_dir = temp_file_path.parent
                if not parent_dir.exists():
                    os.makedirs(str(parent_dir), mode=0o700)

                if content is not None:
                    # create file locally, as needed to push to remote repository
                    temp_file_path.write_bytes(content)
                else:
                    shutil.copy(str(local_path), str(temp_file_path))

                git_repo.git.add(file_path)

                # build commit message
                commit_message = 'File update {name}'.format(name=name)

                if self.tracking_record_id is not None:
                    commit_message += ' for tracking record {}'.format(self.tracking_record_id)

                # commit locally
                # allow-empty will still add commit message with new information even if data itself did not change
                git_repo.git.commit('--allow-empty', '-m', commit_message)

                # push changes remotely
                git_repo.git.push()

                logger.info("File {name} successfully saved".format(name=name))

            except git.exc.GitCommandError:
                logger.error("Unable to save file {name} to remote server.".format(name=name))
                raise

    def retrieve_data(self, name, **kwargs):
        with util.TempDir() as temp_folder:
            file_path, temp_file_path, repo = self.__clone(folder=temp_folder.path, name=name)
            if temp_file_path.is_file():
                return temp_file_path.read_bytes()
            else:
                raise exception.NotFound("File does not exist: {file_path}".format(file_path=file_path))

    def __clone(self, folder, name):
        # clone repository in local folder
        git_repo = git.Repo.clone_from(self.git_url, folder)

        # set user identity
        if self.config_user_name:
            git_repo.config_writer().set_value("user", "name", self.config_user_name).release()
        if self.config_user_email:
            git_repo.config_writer().set_value("user", "email", self.config_user_email).release()

        file_path = name
        if self.folder is not None:
            file_path = self.folder / name

        temp_file_path = folder / file_path

        return file_path, temp_file_path, git_repo



try:
    import unittest.mock as mock
except ImportError:
    import mock

import pytest

from autossl import credential, util

from tests import util as tests_util


def test_get_credentials_already_retrieved_once():
    username = tests_util.get_random_ascii_string(10)
    password = tests_util.get_random_ascii_string(10)
    user_password_credential = credential.get_credentials(
        name='credential_1',
        global_config=None,
        credentials={'credential_1': {
            'type': credential.CredentialType.UserPassword,
            'username': username,
            'password': password,
        }},
        extra_parameters=None,
    )
    assert user_password_credential == {
        'type': credential.CredentialType.UserPassword,
        'username': username,
        'password': password,
    }


def test_get_credentials_unsupported_credential_type():
    with pytest.raises(KeyError):
        credential.get_credentials(
            name='credential_1',
            global_config={'credential_1': {'type': 'not_supported_type'}},
            credentials={},
            extra_parameters=None,
        )


def test_get_credentials_with_extra_parameters():
    # username/password
    assert credential.get_credentials(
        name='credential_1',
        global_config={'credential_1': {'type': credential.CredentialType.UserPassword.name}},
        credentials={'credential_1': {'raw_content': 'my_username:my_pa*ss:wor$d'}},
        extra_parameters={'domain': 'corp'},
    ) == {
        'type': credential.CredentialType.UserPassword,
        'username': 'my_username',
        'password': 'my_pa*ss:wor$d',
        'domain': 'corp',
    }

    # check apiKey credential type with custom separator
    assert credential.get_credentials(
        name='api_credential',
        global_config={'api_credential': {'type': credential.CredentialType.ApiKeyAndId.name, 'separator': '/'}},
        credentials={'api_credential': {'raw_content': '499ebd66-29d6-4992-9ec4-5511a92d248e/12345'}},
        extra_parameters={'additional_identifier': 'corp_789'},
    ) == {
        'type': credential.CredentialType.ApiKeyAndId,
        'api_key': '499ebd66-29d6-4992-9ec4-5511a92d248e',
        'api_id': '12345',
        'additional_identifier': 'corp_789',
    }


def test_get_user_password_from_user_input():
    username = tests_util.get_random_ascii_string(10)
    password = tests_util.get_random_ascii_string(10)

    mock_input = '__builtin__.raw_input' if util.PY2 else 'builtins.input'
    with mock.patch(mock_input, return_value=username), mock.patch('getpass.getpass', return_value=password):
        user_password_credential = credential.get_user_password(name='credential_1')
    assert user_password_credential == {
        'type': credential.CredentialType.UserPassword,
        'username': username,
        'password': password,
    }


@pytest.mark.parametrize('credentials_dict, username, password', [
    ({'credential_1': {'username': 'my_username', 'password': 'my_password'}}, 'my_username', 'my_password'),
    ({'credential_1': {'raw_content': 'my_username:my_pa*ss:wor$d'}}, 'my_username', 'my_pa*ss:wor$d'),
])
def test_get_user_password_from_credentials_list(credentials_dict, username, password):
    user_password_credential = credential.get_user_password(name='credential_1', credentials=credentials_dict)
    assert user_password_credential == {
        'type': credential.CredentialType.UserPassword,
        'username': username,
        'password': password,
    }


def test_get_credentials_from_env_variables(monkeypatch):
    # expected credentials
    username = 'my_username_1'
    password = 'my_pa*ss:wor$d_+'

    # temporarily set env variable with credentials
    monkeypatch.setenv('CREDENTIAL_1_USERNAME', username)
    monkeypatch.setenv('CREDENTIAL_1_PASSWORD', password)

    assert credential.get_credentials(
        name='credential_1',
        global_config={'credential_1': {'type': credential.CredentialType.UserPassword.name}},
        credentials={},
    ) == {
        'type': credential.CredentialType.UserPassword,
        'username': username,
        'password': password,
    }



# standard packages
import os
import shutil
import tempfile

# external packages
import pytest

# autossl imports
from autossl import exception, manager, ssl, ca_manager, util
from autossl.ca_manager import acme_v2_http01

from tests import util as tests_util


@pytest.fixture(scope='function')
def blueprint_name():
    return 'local.example.com.yaml'


@pytest.fixture(scope='function')
def ssl_manager(blueprint_name):
    # create temporary directories use by blueprint
    temp_crt_dir = tempfile.mkdtemp()
    os.environ['AUTOSSL_CRT_PATH'] = temp_crt_dir
    temp_crt_dir_2 = tempfile.mkdtemp()
    os.environ['AUTOSSL_CRT_PATH_2'] = temp_crt_dir_2

    temp_storage_dir = tempfile.mkdtemp()
    os.environ['AUTOSSL_STORAGE_PATH'] = temp_storage_dir
    temp_tracking_dir = tempfile.mkdtemp()
    os.environ['AUTOSSL_TRACKING_PATH'] = temp_tracking_dir

    yield manager.SslManager(
        global_config=None,
        blueprint_path=tests_util.DATA_PATH / blueprint_name,
        credentials=None,
        staging=True
    )

    # cleanup generated artifacts
    shutil.rmtree(temp_crt_dir, ignore_errors=True)
    shutil.rmtree(temp_crt_dir_2, ignore_errors=True)
    shutil.rmtree(temp_storage_dir, ignore_errors=True)
    shutil.rmtree(temp_tracking_dir, ignore_errors=True)


@pytest.mark.parametrize('blueprint_name, api_class, renewal_supported', [
    ('local.example.com.yaml', ca_manager.base.CaManager, False),
    ('acme.example.com.yaml', acme_v2_http01.AcmeHttp01, True),
])
def test_get_ca_manager_api(ssl_manager, api_class, renewal_supported):
    api = ssl_manager.get_ca_manager_api()
    assert type(api) is api_class
    assert api.is_automated_renewal_supported is renewal_supported


def test_get_api_errors():
    ssl_manager = manager.SslManager()
    with pytest.raises(RuntimeError):
        ssl_manager.get_tracking_api()
    with pytest.raises(RuntimeError):
        ssl_manager.get_storage_api()
    with pytest.raises(RuntimeError):
        ssl_manager.get_ca_storage_api()
    with pytest.raises(RuntimeError):
        ssl_manager.get_server_api(server_parameters=None)


@pytest.mark.parametrize('blueprint_name, is_default_storage', [
    ('local.example.com.yaml', True),
    ('acme.example.com.yaml', False),
])
def test_get_ca_storage_api(ssl_manager, is_default_storage):
    # check where CA storage API is same as default storage or dedicated storage instance
    assert (ssl_manager.get_ca_storage_api() is ssl_manager.get_storage_api()) is is_default_storage


def test_get_renewal_status(ssl_manager):
    to_be_renewed, servers_to_update = ssl_manager.get_renewal_status()
    assert to_be_renewed is True
    assert len(servers_to_update) == 2


def test_request_renewal(ssl_manager):
    # make sure folder is empty at beginning
    assert len(os.listdir(os.environ['AUTOSSL_STORAGE_PATH'])) == 0

    # requesting renewal should save files specified in blueprint in storage: csr, private key, ...
    ssl_manager.request_renewal(
        force=True,  # disable interactive user input
    )

    # check files are saved in storage (the one specified in blueprint)
    assert len(os.listdir(os.environ['AUTOSSL_STORAGE_PATH'])) == 2
    # both key and csr should be saved
    for filename in os.listdir(os.environ['AUTOSSL_STORAGE_PATH']):
        assert filename.rsplit('.', 1)[-1] in ['key', 'csr']


def test_renew_certificate(ssl_manager):
    # should do nothing as automated renewal not supported
    ssl_manager.renew_certificate()


@pytest.mark.parametrize('artifacts_from', ['storage', 'input'])
def test_get_and_check_artifacts_ok(tmp_path, ssl_manager, artifacts_from):
    key_path, crt_path = tests_util.create_autossl_signed_cert(
        ssl_blueprint=ssl_manager.ssl_blueprint,
        output_folder=os.environ['AUTOSSL_STORAGE_PATH'],
    )

    parameters = {
        'folder': str(tmp_path)
    }
    if artifacts_from == 'input':
        parameters.update({
            'certificate_path': crt_path,
            'private_key_path': key_path,
        })

    certificate_path, private_key_path = ssl_manager.get_and_check_artifacts(**parameters)
    assert key_path.read_bytes() == private_key_path.read_bytes()
    assert crt_path.read_bytes() == certificate_path.read_bytes()


def test_get_and_check_artifacts_error(ssl_manager):
    # generate certificates and keys directly in storage path
    key_path, crt_path = tests_util.create_self_signed_certificate(
        crt_name=ssl_manager.ssl_blueprint.name,
        output_path=os.environ['AUTOSSL_STORAGE_PATH'],
        common_name=ssl_manager.ssl_blueprint.certificate.common_name,
        sans=ssl_manager.ssl_blueprint.certificate.sans,
        certificate_validity_days=0,
    )
    key_path_2, crt_path_2 = tests_util.create_self_signed_certificate(
        crt_name=ssl_manager.ssl_blueprint.name + '_2',
        output_path=os.environ['AUTOSSL_STORAGE_PATH'],
        common_name=ssl_manager.ssl_blueprint.certificate.common_name,
        sans=ssl_manager.ssl_blueprint.certificate.sans,
    )

    # test for incompatibility between private key and cert
    with pytest.raises(exception.KeyMismatch) as ex_info:
        ssl_manager.get_and_check_artifacts(
            certificate_path=crt_path,
            private_key_path=key_path_2,
        )
    assert 'Certificate specified does not match private key' in str(ex_info.value)

    # test for expiration of cert
    with pytest.raises(exception.ExpiredCertificate) as ex_info:
        ssl_manager.get_and_check_artifacts(
            certificate_path=crt_path,
            private_key_path=key_path,
        )
    assert 'already expired or expiring soon, renewal needed' in str(ex_info.value)

    with pytest.raises(exception.InvalidTrustChain) as ex_info:
        ssl_manager.get_and_check_artifacts(
            certificate_path=crt_path_2,
            private_key_path=key_path_2,
        )
    assert 'does not match CA certificate specified' in str(ex_info.value)


@pytest.mark.parametrize('blueprint_name', ['local_no_chain_of_trust.example.com.yaml'])
def test_get_and_check_artifacts_error_missing_chain_of_trust(ssl_manager):
    # check missing chain of trust in blueprint when full_chain_deploy explicitly requested for a server
    key_path, crt_path = tests_util.create_self_signed_certificate(
        crt_name=ssl_manager.ssl_blueprint.name,
        output_path=os.environ['AUTOSSL_STORAGE_PATH'],
        common_name=ssl_manager.ssl_blueprint.certificate.common_name,
        sans=ssl_manager.ssl_blueprint.certificate.sans,
    )
    with pytest.raises(exception.SslBlueprintInconsistency) as ex_info:
        ssl_manager.get_and_check_artifacts(
            certificate_path=crt_path,
            private_key_path=key_path,
        )
    assert 'missing chain of trust' in str(ex_info.value)


def test_deploy(ssl_manager):
    key_path, crt_path = tests_util.create_autossl_signed_cert(
        ssl_blueprint=ssl_manager.ssl_blueprint,
        output_folder=os.environ['AUTOSSL_STORAGE_PATH'],
    )

    # deploy them on server
    ssl_manager.deploy(
        certificate_path=crt_path,
        private_key_path=key_path
    )

    to_be_renewed, servers_to_update = ssl_manager.get_renewal_status()
    assert to_be_renewed is False
    assert len(servers_to_update) == 0

    # ensure first server contain only server certificate
    crt_path_server_1 = util.Path(os.environ['AUTOSSL_CRT_PATH']).joinpath(ssl_manager.ssl_blueprint.name + '.crt')
    assert crt_path_server_1.read_text().count('-----BEGIN CERTIFICATE-----') == 1
    assert crt_path_server_1.read_bytes() == crt_path.read_bytes()

    # and 2nd server contains certificate with full chain of trust (2 certificates)
    crt_path_server_2 = util.Path(os.environ['AUTOSSL_CRT_PATH_2']).joinpath(ssl_manager.ssl_blueprint.name + '.crt')
    crt_server_2_content = crt_path_server_2.read_text()
    # we should have 2 certificates: server + CA certificates
    assert crt_server_2_content.count('-----BEGIN CERTIFICATE-----') == 2
    # retrieve server certificate (= the first one) and verify it
    first_certificate = '-----BEGIN CERTIFICATE-----' + crt_server_2_content.split('-----BEGIN CERTIFICATE-----')[1]
    assert first_certificate.strip() == crt_path.read_text().strip()


@pytest.mark.parametrize('api_names', [
    None,  # all default apis
    ['tracking'],
    ['storage'],
    ['tracking', 'storage'],
])
def test_save_and_get_file(tmp_path, ssl_manager, api_names):
    with pytest.raises(exception.NotFound):
        ssl_manager.get_file(
            file_type=ssl.DataType.Certificate,
            file_identifier='dummy',
            output_folder=tmp_path,
        )

    file_content = b'some content'
    ssl_manager.save_file(
        file_type=ssl.DataType.Certificate,
        file_content=file_content,
        api_names=api_names,
    )

    file_identifier = '{}.{}'.format(ssl_manager.ssl_blueprint.name, ssl.DataType.Certificate.value)

    # if specific api specified, ensure file is not present in others
    if api_names:
        # remove input apis from default list
        other_apis = list(set(manager.DEFAULT_APIS) - set(api_names))

        # targeting api where file is not saved raise NotFound exception
        with pytest.raises(exception.NotFound):
            ssl_manager.get_file(
                file_type=ssl.DataType.Certificate,
                file_identifier=file_identifier,
                output_folder=tmp_path,
                api_names=other_apis,
            )

    # here we should find file as same apis than for 'save_file' are used
    file_path_result = ssl_manager.get_file(
        file_type=ssl.DataType.Certificate,
        file_identifier=file_identifier,
        output_folder=tmp_path,
        api_names=api_names,
    )
    assert file_path_result.read_bytes() == file_content


@pytest.mark.parametrize('blueprint_name', ['localca.example.com.yaml'])
def test_automatic_renewal(ssl_manager):
    shutil.copy(str(tests_util.DATA_PATH / 'ca' / 'autossl_ca.key'), os.environ['AUTOSSL_STORAGE_PATH'])
    shutil.copy(str(tests_util.DATA_PATH / 'ca' / 'autossl_ca.crt'), os.environ['AUTOSSL_STORAGE_PATH'])

    # create csr/private key
    ssl_manager.request_renewal(
        force=True,  # disable interactive user input
    )

    # sign certificate with local CA
    ssl_manager.renew_certificate()

    # verify certificate validity
    assert os.system('openssl verify -CAfile %s %s' % (
        tests_util.DATA_PATH / 'ca' / 'autossl_ca.crt',
        os.path.join(os.environ['AUTOSSL_STORAGE_PATH'], 'localca.example.com.crt')
    )) == 0




class Storage(object):

    def __init__(self, tracking_record_id=None, **kwargs):
        r"""Base interface to store data

        :param tracking_record_id: identified of tracking record if tracking system is used
        :type tracking_record_id:
        :param \**kwargs: key/value parameters needed for initialization
        :type \**kwargs: dict
        """
        self.tracking_record_id = tracking_record_id

    def save_data(self, name, data_type, content=None, local_path=None, **kwargs):
        r"""Save specified content in storage

        :param name: name of the content to be stored on server side
        :type name: str
        :param data_type: type of data to save
        :type data_type: ssl.DataType
        :param content: content to be stored on server side
        :type content: bytes
        :param local_path: local path to a file to store
        :type local_path: pathlib.Path or str
        :param \**kwargs: optional key/value parameters from blueprint to save data
        :type \**kwargs: dict

        Either one of `content` or `local_path` must be specified but not both
        """
        raise NotImplementedError("Must be overridden in storage specific implementation.")

    def retrieve_data(self, name, data_type, **kwargs):
        r"""Retrieve data from storage

        :param name: identifier of data to retrieve
        :type name: str
        :param data_type: type of data to retrieve
        :type data_type: ssl.DataType
        :param \**kwargs: optional key/value parameters from blueprint to retrieve data
        :type \**kwargs: dict
        :return: requested data
        :rtype: bytes
        :raise exception.NotFound: when requested data are missing in storage
        """
        raise NotImplementedError("Must be overridden in storage specific implementation.")



import importlib
import logging
try:
    # Python 3
    from pathlib import Path  # noqa: F401
except ImportError:
    #  Python 2
    from pathlib2 import Path  # noqa: F401
import requests
import shutil
from six import PY2  # noqa: F401
import tempfile

from . import exception

logger = logging.getLogger(__name__)


def check_http_response_ok(response):
    """Validate http response code

    all codes not in 2xx will raise an exception

    :param response: requests Http response
    :type response: requests.Response
    :return: same http response
    :rtype: requests.Response
    :raise exception.HttpCodeException: if http status code in not in 2xx
    """
    try:
        response.raise_for_status()
    except requests.exceptions.HTTPError as ex:
        raise exception.HttpCodeException(ex)
    return response


def str_to_class(class_path):
    """Dynamically import and return class type from full module and class path

    :param class_path:
    :type class_path: str
    :return: Type of the class to instantiate
    :rtype: type
    :raise ImportError: if module does not exist
    :raise AttributeError: if class not found in specified module
    """
    module_name, class_name = class_path.rsplit('.', 1)
    try:
        module_ = importlib.import_module(module_name)
        try:
            return getattr(module_, class_name)
        except AttributeError:
            logging.exception('Class %s not found in module %s.' % (class_name, module_name))
            raise
    except ImportError:
        logging.exception('Module %s does not exist.' % module_name)
        raise


class TempDir(object):
    def __init__(self, path=None):
        """Create Temporary directory that can be used with context manager for automated deletion at __exit__
        :param path: local path. If None, temporary folder will be created at `__enter__` thanks to `tempfile.mkdtemp()`
        :type path: str or pathlib.Path
        """
        self.input_path = path
        self.path = None

    def __enter__(self):
        self.path = Path(str(self.input_path) if self.input_path else tempfile.mkdtemp())
        if not self.path.is_dir():
            raise IOError("Specified path {} is not a directory.".format(self.path))
        return self

    def __exit__(self, *args):
        if self.path.exists() and self.path.is_dir():
            shutil.rmtree(str(self.path), ignore_errors=True)



"""
Tests for Git storage implementation
"""
import pytest
import tempfile

import git

from autossl import exception
from autossl.storage import gitscm

from tests import util as tests_util


@pytest.fixture(scope="function")
def folder_name():
    return None


@pytest.fixture(scope="function")
def tracking_record_id():
    return None


@pytest.fixture(scope="function")
def git_storage(tmp_path, folder_name, tracking_record_id):
    # initialize bare git repository in temporary folder
    git.Repo.init(str(tmp_path), bare=True)

    # return git storage instance
    yield gitscm.GitStorage(git_url=str(tmp_path), folder=folder_name, tracking_record_id=tracking_record_id,
                            config_user_name='Test User', config_user_email='test.user@autossl.com')


def test_retrieve_data_error(git_storage):
    with pytest.raises(exception.NotFound):
        git_storage.retrieve_data(name='my_file_name')


@pytest.mark.parametrize('folder_name,tracking_record_id', [
    (None, None),
    ('first_folder', 123456789)
])
def test_save_retrieve_data_from_content(git_storage, folder_name, tracking_record_id):
    file_name = tests_util.get_random_ascii_string()
    file_content = tests_util.get_random_text()
    # save content in main repo
    git_storage.save_data(name=file_name, content=file_content)

    assert git_storage.retrieve_data(name=file_name) == file_content


@pytest.mark.parametrize('folder_name', [
    None,
    'other_folder'
])
def test_save_retrieve_data_from_local_path(git_storage, folder_name):
    file_name = tests_util.get_random_ascii_string()
    file_content = tests_util.get_random_text()

    # save content in main repo from local_path
    with tempfile.NamedTemporaryFile() as local_file:
        local_file.write(file_content)
        local_file.seek(0)
        git_storage.save_data(name=file_name, local_path=local_file.name)

    assert git_storage.retrieve_data(name=file_name) == file_content


def test_save_retrieve_data_same_file_twice(git_storage):
    file_name = tests_util.get_random_ascii_string()
    file_content = tests_util.get_random_text()

    # save content in main repo
    git_storage.save_data(name=file_name, content=file_content)

    assert git_storage.retrieve_data(name=file_name) == file_content

    # try to save same data a 2nd time should not fail
    git_storage.save_data(name=file_name, content=file_content)

    assert git_storage.retrieve_data(name=file_name) == file_content


def test_git_url_with_username_password_present():
    git_url = 'https://test/heya.git'
    username = 'Jack'
    password = 'thisisaclearpassword,stopreading'
    new_git_url = gitscm.git_url_with_username_password(git_url, username, password)
    assert new_git_url == git_url.replace('://', '://{}:{}@'.format(username, password))


def test_git_url_with_username_password_absent():
    git_url = 'https://test/heya.git'
    new_git_url = gitscm.git_url_with_username_password(git_url, None, None)
    assert git_url == new_git_url



class Server(object):

    def __init__(self, crt_name, deploy_full_chain=False, **kwargs):
        r"""Default server implementation describing interface of a server
        This is an abstract class, so each specialized method must be overridden in parent class.

        :param crt_name: name of certificate on server (without any extension like .crt)
        :type crt_name: str
        :param deploy_full_chain: if True, deploy server certificate with full chain of trust in crt file
            else, simply deploy server certificate in crt file
        :type deploy_full_chain: bool
        """
        self.crt_name = crt_name
        self.deploy_full_chain = deploy_full_chain

    def get_description(self):
        """Get description of this server

        :return: server description
        :rtype: str
        """
        raise NotImplementedError("Must be overridden in parent class")

    def deploy_cert(self, key, cert, **kwargs):
        r"""Deploy input certificate on server

        :param key: path to local private key
        :type key: pathlib.Path
        :param cert: path to local public certificate
        :type cert: pathlib.Path
        :raise exception.DeployCertificateError: if unexpected error occurred during deployment on server
        """
        raise NotImplementedError("Must be overridden in parent class")

    def create_acme_challenge(self, token, key_authorization):
        """Create token on server with specified value

        :param token: challenge key
        :param key_authorization: challenge value
        """
        raise NotImplementedError("Must be overridden in parent class")

    def delete_acme_challenge(self, token):
        """Delete challenge created on server

        :param token: challenge key to delete from server
        :type token: str
        """
        raise NotImplementedError("Must be overridden in parent class")

    def get_certificate_information(self):
        """Retrieve certificate information from server.

        Must be implemented for each type of server.

        :return: SSL certificate information
        :rtype: autossl.ssl.SslCertificate
        :raise autossl.exception.CertificateNotFound: if certificate does not exist yet on server
        """
        raise NotImplementedError("Must be overridden in parent class")

    def is_same(self, common_name=None, sans=None, exact_match=False):
        """Check if current certificate deployed on server is covering all specified domains

        :param common_name: Common name
        :type common_name: str
        :param sans: list of Subject Alternate Names
        :type sans: list
        :param exact_match: if True, certificate must exactly match input domains
            if False, input domain will also match wilcard certificate and additional domains in certificate will
            be ignored
        :type exact_match: bool
        :return: True is certificate is already covering all domains
        """
        return self.get_certificate_information().is_same(common_name, sans, exact_match)

    def is_expired(self, expiration_delay=0):
        """Check for expiration of specified certificate

        :param expiration_delay: Number of days before real expiration we consider a renewal needed
        :type expiration_delay: int
        :return: True is certificate is going to expire in less than expiration_delay days
        :rtype: bool
         """
        return self.get_certificate_information().is_expired(expiration_delay)



import pytest

from cryptography.hazmat.backends import default_backend
from cryptography.x509.oid import NameOID
from cryptography import x509

from autossl.ca_manager import local as local_ca
from autossl.storage import local as local_storage
from autossl import ssl
from tests import util as tests_util


@pytest.fixture(scope="function")
def storage(tmp_path):
    yield local_storage.LocalFileStorage(path=str(tmp_path))


@pytest.fixture(scope="function")
def ca_manager(storage):
    ca_key, ca_crt = tests_util.create_ca_certificate(ca_name='Autossl')
    storage.save_data(name='ca_key', data_type=ssl.DataType.PrivateKey, content=ca_key)
    storage.save_data(name='ca_crt', data_type=ssl.DataType.Certificate, content=ca_crt)
    yield local_ca.LocalCa(ca_config=None,
                           staging=True,
                           storage_api=storage,
                           ca_private_key='ca_key',
                           ca_certificate='ca_crt')


def test_automated_renewal_supported(ca_manager):
    assert ca_manager.is_automated_renewal_supported is True


@pytest.mark.parametrize('common_name', ['test.autossl.com'])
def test_get_signed_certificate(ca_manager, common_name, tmp_path):
    _, csr_path = ssl.generate_csr(name='autossl_cert',
                                   common_name=common_name,
                                   output_path=tmp_path)
    crt = ca_manager.get_signed_certificate(ssl_blueprint=None, csr_path=csr_path, servers_api=None)

    # check CRT
    x509_object = x509.load_pem_x509_certificate(data=crt, backend=default_backend())
    assert len(x509_object.subject.get_attributes_for_oid(NameOID.COMMON_NAME)) == 1
    assert x509_object.subject.get_attributes_for_oid(NameOID.COMMON_NAME)[0].value == common_name



import logging

import pytest

import autossl
from autossl import util, __main__

from tests import util as tests_util


def test_display_version():
    assert __main__.display_version() == "Autossl version {version}".format(version=autossl.__version__)


@pytest.mark.parametrize('cli_credentials, credentials_file, result', [
    # full user_password credentials from cli, no credential file
    (['credential_1:my_username:my_pa*ss:wor$d'], None,
     {'credential_1': {'raw_content': 'my_username:my_pa*ss:wor$d'}}),
    # no password from cli, no credential file
    (['credential_1:my_username'], None,
     {'credential_1': {'raw_content': 'my_username'}}),
    # just credential name from cli, no credential file
    (['credential_1'], None,
     {'credential_1': {'raw_content': None}}),
    # full user_password credentials from cli, credential file with cli overriding file for credential_1
    (['credential_1:my_username:my_pa*ss:wor$d'], tests_util.DATA_PATH / 'credentials_file',
     {'credential_1': {'raw_content': 'my_username:my_pa*ss:wor$d'},
      'credential_4_from_file': {'username': 'username_4', 'password': 'password_4', 'extra_parameter': 'extra_param'},
      'credential_5_from_file': {'api_key': '499ebd66-29d6-4992-9ec4-5511a92d248e', 'api_id': '12345'}}),
])
def test_parse_credentials(cli_credentials, credentials_file, result):
    assert __main__.parse_credentials(cli_credentials=cli_credentials, credentials_file=credentials_file) == result


def test_main_parser(capsys):
    # check action is mandatory
    with pytest.raises(SystemExit):
        __main__.parse_arguments([])
    if util.PY2:
        assert 'error: too few arguments' in capsys.readouterr().err
    else:
        assert 'error: the following arguments are required: action' in capsys.readouterr().err

    # default values (as action is mandatory, use 'version' that requires no additional parameter)
    parser = __main__.parse_arguments(['version'])
    assert parser.credentials == []
    assert parser.debug is logging.INFO
    assert parser.staging is False
    assert parser.config is None
    assert parser.blueprint is None

    # parser = __main__.parse_arguments(['-u', r'domain\user:password', '--debug', 'version'])
    # assert parser.credentials == r'domain\user:password'
    # assert parser.debug is True

# TODO



import datetime
from enum import Enum
import logging
import os
import re
import tempfile
import uuid

# external packages
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography import x509
from cryptography.x509.extensions import ExtensionNotFound
from cryptography.x509.oid import NameOID, ExtensionOID
from cryptography.hazmat.primitives import hashes
import yaml

from . import credential, exception, util

logger = logging.getLogger(__name__)


class DataType(Enum):
    """list of data types supported"""
    CertificateSigningRequest = 'csr'
    PrivateKey = 'key'
    Certificate = 'crt'
    Blueprint = 'yaml'


class SslBlueprint(object):
    def __init__(self, yaml_path=None, global_config_path=None):
        """Blueprint representation of an SSL certificate

        :param yaml_path: path to the yaml file representing the blueprint
        :type yaml_path: pathlib.Path or str
        :param global_config_path: path to the yaml file representing the global configuration
        :type global_config_path: pathlib.Path
        """
        blueprint_content = dict()
        self.blueprint_path = None

        if yaml_path:
            # also support str path in input
            if not isinstance(yaml_path, util.Path):
                yaml_path = util.Path(yaml_path)
            with yaml_path.open() as blueprint_file:
                blueprint_content = yaml.safe_load(blueprint_file)
            self.blueprint_path = yaml_path

        if global_config_path:
            # also support str path in input
            if not isinstance(global_config_path, util.Path):
                global_config_path = util.Path(global_config_path)
            with global_config_path.open() as global_config_path_file:
                global_config_content = yaml.safe_load(global_config_path_file)

            # all keys present in global config and not in blueprint will be added
            for key in global_config_content:
                if key not in blueprint_content:
                    blueprint_content[key] = global_config_content[key]

            # create local temporary file with global config
            with tempfile.NamedTemporaryFile(delete=False, mode='w+') as consolidated_blueprint_file:
                self.consolidated_blueprint_path = util.Path(consolidated_blueprint_file.name)
                yaml.dump(data=blueprint_content, stream=consolidated_blueprint_file)
        else:
            self.consolidated_blueprint_path = self.blueprint_path

        # ## mandatory fields
        # name of certificate on server
        self.name = blueprint_content.get('name')

        # list of servers where certificate must be deployed
        self.servers = blueprint_content.get('servers', [])

        #########################
        # certificate information
        #########################
        self.certificate = None
        self.ca_config = None
        if blueprint_content.get('certificate'):
            self.certificate = SslCertificateConfig(
                # Certificate type
                certificate_type=blueprint_content['certificate'].get('type'),
                # which CA to use for renewal
                certificate_authority=blueprint_content['certificate'].get('certificate_authority'),
            )
            # to specify if certificate on server must exactly match the content of the blueprint
            self.certificate.set_attr_if_not_none(
                'exact_match',
                blueprint_content['certificate'].get('exact_match')
            )

            # force reuse of existing private key if available in storage, else generate a new one
            self.certificate.set_attr_if_not_none(
                'private_key_reuse',
                blueprint_content['certificate'].get('private_key_reuse')
            )

            self.certificate.set_attr_if_not_none(
                'private_key_size',
                blueprint_content['certificate'].get('private_key_size')
            )

            # Organization information
            # get first information from blueprint directly from certificate blueprint
            # or get it from global config
            self.certificate.set_attr_if_not_none(
                'organization',
                blueprint_content['certificate'].get('organization') or blueprint_content.get('organization')
            )

            # Common Name field of the certificate
            self.certificate.set_attr_if_not_none(
                'common_name',
                blueprint_content['certificate'].get('common_name')
            )

            # List of SAN for this certificate
            self.certificate.set_attr_if_not_none(
                'sans',
                blueprint_content['certificate'].get('san')
            )

            # renew certificate if less than `renewal_delay` days before expiration
            self.certificate.set_attr_if_not_none(
                'renewal_delay',
                blueprint_content['certificate'].get('renewal_delay')
            )

            # chain_of_trust associated to server certificate
            self.certificate.set_attr_if_not_none(
                'chain_of_trust',
                blueprint_content['certificate'].get('chain_of_trust')
            )

            # this certificate will be use to sign other certificates (CA)
            self.certificate.set_attr_if_not_none(
                'is_ca',
                blueprint_content['certificate'].get('is_ca')
            )

            self.ca_config = CertificateAuthorityConfig(
                certificate_authorities=blueprint_content.get('certificate_authorities'),
                certificate_authority_key=self.certificate.certificate_authority,
            )

        # retrieve storage and tracking server configuration
        self.storage = blueprint_content.get('storage')
        self.tracking = blueprint_content.get('tracking')
        self.credentials = blueprint_content.get('credentials')

        # perform extended validation of blueprint content
        self.validate()

    def __del__(self):
        # cleanup temporarily generated consolidated config file if present
        if hasattr(self, 'consolidated_blueprint_path') \
                and hasattr(self, 'blueprint_path') \
                and self.consolidated_blueprint_path.is_file() \
                and self.consolidated_blueprint_path != self.blueprint_path:
            self.consolidated_blueprint_path.unlink()

    def validate(self):
        """Validate data extracted from blueprint

        :raise ValueError: if content of specified blueprint is not valid
        """
        # check certificate information
        if self.certificate:
            self.certificate.validate(ca_config=self.ca_config)

        # validate server config
        for server in self.servers:
            server_credentials_name = server.get('credentials')
            if server_credentials_name is not None:
                # check all server credentials are referenced in global section
                if server_credentials_name not in self.credentials:
                    raise ValueError("Missing credentials {} in global credentials section"
                                     .format(server_credentials_name))

                # check 'type' attribute is present and with a valid value in global credentials section
                credential_type = self.credentials[server_credentials_name].get('type')
                try:
                    credential.CredentialType[credential_type]
                except KeyError:
                    raise ValueError("Credential type {} not supported, expected one of {}"
                                     .format(credential_type, str(credential.CredentialType.__members__.keys())))

    @property
    def domains(self):
        """Get domains covered by this blueprint

        :return: list of domains in blueprint
        :rtype: set
        """
        return self.certificate.domains

    def get_config(self, name, path=None, default=None):
        result = getattr(self, name)
        if path is not None and any(path):
            # no config specified, use base one
            if result is None:
                if len(path) == 1 and path[0] == 'type':
                    # just use base implementation
                    return 'base.%s'.format(name.title())
            else:
                for key in path:
                    if result is None:
                        break
                    result = result.get(key)
        if result is None:
            result = default
        return result

    def get_chain_of_trust(self):
        """Return list of certificates to have full chain of trust: intermediate, root
        :return: list of certificates starting intermediate until root certificate
        :rtype: list
        """
        # chain of trust can be define directly in certificate config
        # or globally at CA config level
        return self.certificate.chain_of_trust or self.ca_config.get_chain_of_trust()


class SslCertificateConfig(object):
    def __init__(self, certificate_type, certificate_authority,
                 common_name=None, sans=None, organization=None, chain_of_trust=None,
                 exact_match=False, private_key_reuse=True, private_key_size=4096,
                 renewal_delay=30, is_ca=False):
        """

        :param certificate_type: ssl certificate type (DV, OV, EV)
        :param certificate_authority: certificate authority name
        :param common_name: ssl certificate common name
        :param sans: ssl certificate alternative name
        :param organization: certificate organization
        :param chain_of_trust: list of certificates to have full chain of trust: intermediate, root
        :param exact_match: to specify if certificate on server must exactly match the content of the blueprint
        :param private_key_reuse: force reuse of existing private key if available in storage, else generate a new one
        :param private_key_size: size of private key if new one must be generated
        :param renewal_delay: renew certificate if less than `renewal_delay` days before expiration
        :param is_ca: this certificate is used to sign other certificates (CA)
        """
        # Mandatory fields
        self.certificate_type = certificate_type
        self.certificate_authority = certificate_authority

        # Optional fields
        self.common_name = common_name
        self.sans = sans or []
        self.organization = organization
        self.chain_of_trust = chain_of_trust or []

        self.exact_match = exact_match
        self.private_key_reuse = private_key_reuse
        self.private_key_size = private_key_size
        self.renewal_delay = renewal_delay
        self.is_ca = is_ca

    @property
    def domains(self):
        # build unique list of domains that is expected to be covered
        return get_domains(common_name=self.common_name, sans=self.sans)

    def set_attr_if_not_none(self, attr_name, value):
        """Set attribute value if value is not None

        :param attr_name: attribute name
        :type attr_name: str
        :param value: attribute value
        """
        if value is not None:
            setattr(self, attr_name, value)

    def validate(self, ca_config):
        if self.certificate_type != 'DV':
            # mandatory for non DV certificates
            if self.organization is None:
                raise ValueError('Missing mandatory organization information for certificate type %s'
                                 % self.certificate_type)

        # a certificate must cover at list 1 domain, so we expect either common name or san or both
        if self.common_name is None and len(self.sans) == 0:
            raise ValueError("At least 1 of attributes 'common_name' or 'san' must be provided")

        # validate format of domains specified
        domain_pattern = r'^(\*\.)?[-\.\w+]{1,255}$'
        if self.common_name:
            if not re.match(domain_pattern, self.common_name):
                raise ValueError("Specified common name '{}' does not match pattern '{}'"
                                 .format(self.common_name, domain_pattern))

        for san in self.sans:
            if not re.match(domain_pattern, san):
                raise ValueError("Specified san '%s' does not match pattern '%s'" % (san, domain_pattern))

        if not isinstance(self.renewal_delay, int):
            raise ValueError("Renewal delay must be an integer got %s" % type(self.renewal_delay))

        if ca_config and not ca_config.is_certificate_supported(cert_type=self.certificate_type):
            raise ValueError('Certificate type %s not supported. Expect one of [%s].'
                             % (self.certificate_type,
                                ', '.join(ca_config.get_supported_certificate_types())))


class CertificateAuthorityConfig(object):
    def __init__(self, certificate_authorities, certificate_authority_key):
        """Config representing CA configuration"""
        self.ca_config = dict()

        # find config for specified CA or raise an exception
        if certificate_authorities is not None:
            for ca_config in certificate_authorities:
                if ca_config.get('key') == certificate_authority_key:
                    self.ca_config = ca_config
                    break
            else:
                raise ValueError("Certificate authority '%s' is not supported. Expect one of [%s]."
                                 % (certificate_authority_key,
                                    ', '.join([ca_config.get('key') for ca_config in certificate_authorities])))

    def get_supported_certificate_types(self):
        """Get list of certificate types currently supported by CA

        :return: list of certificate types currently supported by CA
        :rtype: list
        """
        # if not restricted explicitly, support all certificates types by default
        return self.ca_config.get('certificate_types', ['DV', 'OV', 'EV'])

    def is_certificate_supported(self, cert_type):
        """Check if specified certificate type is supported by CA

        :param cert_type: type of certificate to check (ex: DV)
        :type cert_type: str
        :return: True if CA supports this certificate type
        :rtype: bool
        """
        return cert_type in self.get_supported_certificate_types()

    def is_acme_supported(self):
        """Check if CA supports ACME protocol

        :return: True if CA supports ACME protocol
        :rtype: bool
        """
        return any(self.ca_config.get('acme_api', {}))

    def get_acme_api(self, staging=False):
        acme_apis = self.ca_config.get('acme_api', {})
        if staging:
            return acme_apis.get('staging')
        else:
            return acme_apis.get('production')

    def get_storage_config(self):
        """Get configuration of CA storage api

        :return: CA storage configuration
        :rtype: dict
        """
        return self.ca_config.get('storage')

    def get_chain_of_trust(self):
        """Return list of certificate to have full chain of trust: intermediate, root
        :return: list of certificate starting intermediate until root certificate
        :rtype: list
        """
        return self.ca_config.get('chain_of_trust') or []


def generate_csr(name,
                 common_name=None,
                 company_name=None,
                 street_address=None,
                 city=None,
                 state=None,
                 postal_code=None,
                 country_code=None,
                 email_address=None,
                 sans=None,
                 key_content=None,
                 key_size=4096,
                 output_path=None,
                 is_ca=False):
    """Generate a CSR for specified parameters

    if a private key is given, it will be used to generate CSR, else a new one will be created

    :param name: name of file generated (without extension)
    :param common_name: common name
    :param company_name: company name
    :param street_address: company street address
    :param city: company city
    :param state: company state
    :param postal_code: company postal code
    :param country_code: company country
    :param email_address: contact email
    :param sans: list of SANs to be covered
    :param key_content: optional private key content to generate CSR
    :type key_content: byte
    :param key_size: size of private key to generate CSR, if no key in input
    :param output_path: local path where to generate files
    :param is_ca: True if the requested certificate is for a CA
    :return: tuple(key_content, csr_path) with content of private key and path to csr file
    :rtype: tuple(bytes, pathlib.Path)
    """
    output_path = output_path or os.path.curdir
    if not isinstance(output_path, util.Path):
        output_path = util.Path(str(output_path))
    output_path.mkdir(parents=True, exist_ok=True)

    # use existing private key
    if key_content is not None:
        key = serialization.load_pem_private_key(
            data=key_content,
            password=None,
            backend=default_backend(),
        )
        logger.info("Using existing private key.")
    # Generate our key
    else:
        key = rsa.generate_private_key(
            public_exponent=65537,
            key_size=key_size,
            backend=default_backend(),
        )
        # get private key content as we never want to write it on disk for security reasons
        key_content = key.private_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PrivateFormat.TraditionalOpenSSL,
            encryption_algorithm=serialization.NoEncryption(),
        )
        logger.info("New private key created.")

    name_attributes = []

    if common_name is not None:
        name_attributes.append(x509.NameAttribute(NameOID.COMMON_NAME, u"%s" % common_name))
    if company_name is not None:
        name_attributes.append(x509.NameAttribute(NameOID.ORGANIZATION_NAME, u"%s" % company_name))
    if street_address is not None:
        name_attributes.append(x509.NameAttribute(NameOID.STREET_ADDRESS, u"%s" % street_address))
    if postal_code is not None:
        name_attributes.append(x509.NameAttribute(NameOID.POSTAL_CODE, u"%s" % postal_code))
    if state is not None:
        name_attributes.append(x509.NameAttribute(NameOID.STATE_OR_PROVINCE_NAME, u"%s" % state))
    if city is not None:
        name_attributes.append(x509.NameAttribute(NameOID.LOCALITY_NAME, u"%s" % city))
    if country_code is not None:
        name_attributes.append(x509.NameAttribute(NameOID.COUNTRY_NAME, u"%s" % country_code))
    if email_address is not None:
        name_attributes.append(x509.NameAttribute(NameOID.EMAIL_ADDRESS, u"%s" % email_address))

    # Generate a CSR
    builder = x509.CertificateSigningRequestBuilder()
    builder = builder.subject_name(x509.Name(name_attributes))
    builder = builder.add_extension(
        x509.SubjectAlternativeName([x509.DNSName(u"%s" % alt_domain) for alt_domain in sans or []]),
        critical=False
    )
    if is_ca:
        builder = builder.add_extension(
            x509.BasicConstraints(ca=True, path_length=0), critical=True
        )
        builder = builder.add_extension(
            x509.KeyUsage(
                digital_signature=True,
                content_commitment=False,
                key_encipherment=True,
                data_encipherment=False,
                key_agreement=False,
                key_cert_sign=True,
                crl_sign=False,
                encipher_only=False,
                decipher_only=False
                ),
            critical=True,
        )
        if sans:
            builder = builder.add_extension(
                x509.NameConstraints([x509.DNSName(u"%s" % alt_domain) for alt_domain in sans], []), critical=True
            )

    # Sign the CSR with our private key.
    csr = builder.sign(key, hashes.SHA256(), default_backend())

    # Write our CSR out to disk.
    csr_path = output_path.joinpath("%s.csr" % name)
    csr_path.write_bytes(csr.public_bytes(serialization.Encoding.PEM))
    logger.info("Csr %s created." % csr_path.name)

    return key_content, csr_path


def sign(csr, ca_key, ca_cert, validity_days):
    """Sign a certificate request with a key (CA)

    :param csr: certificate request to sign
    :type csr: bytes, PEM encoded
    :param ca_key: the signing key
    :type ca_key: bytes, PEM encoded
    :param ca_cert: the signing certificate
    :type ca_cert: bytes, PEM encoded
    :param validity_days: certificate validity duration (in days)
    :type validity_days: int
    :return: the signed certificate
    :rtype: bytes, PEM encoded
    """
    csr = x509.load_pem_x509_csr(data=csr, backend=default_backend())
    ca_key = serialization.load_pem_private_key(
        data=ca_key,
        password=None,
        backend=default_backend(),
    )
    ca_crt = x509.load_pem_x509_certificate(data=ca_cert, backend=default_backend())

    now = datetime.datetime.utcnow()
    builder = x509.CertificateBuilder() \
        .subject_name(csr.subject) \
        .issuer_name(ca_crt.subject) \
        .public_key(csr.public_key()) \
        .serial_number(uuid.uuid4().int) \
        .not_valid_before(now) \
        .not_valid_after(now + datetime.timedelta(days=validity_days)) \
        .add_extension(extension=x509.AuthorityKeyIdentifier.from_issuer_public_key(ca_key.public_key()),
                       critical=False)
    for extension in csr.extensions:
        builder = builder.add_extension(extension.value, extension.critical)
    crt = builder.sign(
        private_key=ca_key,
        algorithm=hashes.SHA256(),
        backend=default_backend()
    )
    return crt.public_bytes(encoding=serialization.Encoding.PEM)


def get_domains_from_x509(file_path, file_type):
    """Retrieve the list of domains covered by specified x509 file (CSR or CRT)

    :param file_path: path to x509 file
    :type file_path: pathlib.Path
    :param file_type: type of x509 file.
        Supported types: [DataType.CertificateSigningRequest, DataType.Certificate]
    :type file_type: DataType
    :return: list of domain
    :rtype: set
    """
    domains = set()

    # load csr first
    if file_type == DataType.CertificateSigningRequest:
        x509_object = x509.load_pem_x509_csr(data=file_path.read_bytes(), backend=default_backend())
    elif file_type == DataType.Certificate:
        x509_object = x509.load_pem_x509_certificate(data=file_path.read_bytes(), backend=default_backend())
    else:
        raise ValueError("x509 file type specified not supported: ''" % file_type)

    # common_name (optional, we can have only sans)
    common_name = x509_object.subject.get_attributes_for_oid(NameOID.COMMON_NAME)
    if len(common_name) > 0:
        domains.add(common_name[0].value)

    # sans
    try:
        for dns_name in x509_object.extensions.get_extension_for_oid(ExtensionOID.SUBJECT_ALTERNATIVE_NAME).value:
            domains.add(dns_name.value)
    except ExtensionNotFound:
        # no san in that CSR
        pass

    return domains


def get_expiration(crt_path):
    x509_object = x509.load_pem_x509_certificate(data=crt_path.read_bytes(), backend=default_backend())
    return x509_object.not_valid_after


def check_certificate_with_key(key_path, crt_path):
    """Check whether a private key matches a certificate

    For this, we compare RSAPublicNumbers from public key in certificate
    with the RSAPublicNumbers which makes up the RSA public key associated with this RSA private key.

    :param key_path: path to private key
    :type key_path: pathlib.Path
    :param crt_path: path to SSL certificate
    :type crt_path: pathlib.Path
    :return: True, if certificate matches private key
    :rtype: bool
    """
    # load private key
    key = serialization.load_pem_private_key(
        data=key_path.read_bytes(),
        password=None,
        backend=default_backend())

    # load certificate
    certificate = x509.load_pem_x509_certificate(
        data=crt_path.read_bytes(),
        backend=default_backend())

    return key.private_numbers().public_numbers == certificate.public_key().public_numbers()


def check_chain_of_trust(chain_of_trust, crt_path):
    """Check that input certificate matches chain of trust

    :param chain_of_trust: list of certificates of the chain of trust (intermediate CA, root CA)
    :type chain_of_trust: list
    :param crt_path: local path to certificate to verify
    :type crt_path: pathlib.Path
    :raises exception.InvalidTrustChain: if input certificate does not match chain of trust specified
    """
    with util.TempDir() as temp_folder:
        # temporarily write CA information in bundle file
        bundle_path = temp_folder.path.joinpath(str(crt_path) + '.bundle')
        bundle_path.write_bytes('\n'.join(chain_of_trust).encode('utf-8'))

        # verify directly with openssl as unable to find a way with cryptography module
        if os.system('openssl verify -CAfile {} {}'.format(bundle_path, crt_path)) != 0:
            raise exception.InvalidTrustChain(
                "Certificate {} does not match CA certificate specified".format(crt_path.name))


def is_domain_matching(domain_to_check, reference_domain, exact_match=False):
    r"""Check if a domain is matching another domain

    For example, test.example.com is matching by \*.example.com

    :param domain_to_check: the domain to check
    :param reference_domain: the reference domain to compare with
    :param exact_match: If True, domain_to_check and reference_domain must be the same
        If False, domain_to_check can be only a subset of reference_domain
    :return: True if domain_to_check is matching reference_domain
    :rtype: bool
    """
    if domain_to_check == reference_domain:
        return True
    elif not exact_match:
        # use regexp to support wildcard domain
        regex = reference_domain.replace('.', r'\.').replace('*', r'.*') + '$'
        if re.match(regex, domain_to_check):
            return True
    return False


def is_domain_list_matching(domains_to_check, reference_domains, exact_match=False):
    r"""Check if a list of domains are covered by another list of domains

    For example, test.example.com and test2.example.com are covered by \*.example.com

    :param domains_to_check: list of domains to check
    :param reference_domains: list of reference domains to compare with
    :param exact_match: If True, domains_to_check and reference_domains must be the same
        If False, domains_to_check can be only a subset of reference_domains
    :return: True if domains_to_check are covered by reference_domains
    :rtype: bool
    """
    domains_not_covered = set()
    for domain_to_check in domains_to_check:
        for reference_domain in reference_domains:
            if is_domain_matching(domain_to_check, reference_domain, exact_match=exact_match):
                break
        else:
            domains_not_covered.add(domain_to_check)

    if any(domains_not_covered):
        logger.info("Following domains not covered by certificate: [%s]" % (', '.join(domains_not_covered)))
        return False
    logger.debug("Blueprint domains all covered by current certificate")
    return True


def get_domains(common_name=None, sans=None):
    """Get unique list of domains for input criteria

    :param common_name: Certificate common name
    :type common_name: str or None
    :param sans: Certificate SANs List
    :type sans: list(str) or None
    :return: unique list of domains
    :rtype: set(str)
    """
    domains = set()
    if common_name is not None:
        domains.add(common_name)
    if sans:
        domains |= set(sans)
    return domains


class SslCertificate(object):
    def __init__(self, x509_path=None, common_name=None, sans=None, expiration=None):
        self.common_name = common_name
        self.sans = sans or []
        self.expiration = expiration

        if x509_path:
            self.init_from_x509(x509_path)

    def __eq__(self, other):
        return self.common_name == other.common_name \
               and sorted(self.sans) == sorted(other.sans) \
               and self.expiration == other.expiration

    def __ne__(self, other):
        return not self.__eq__(other)

    def init_from_x509(self, x509_path):
        """
        :param x509_path: path to PEM certificate
        :type x509_path: pathlib.Path
        """
        if not isinstance(x509_path, util.Path):
            x509_path = util.Path(str(x509_path))
        x509_object = x509.load_pem_x509_certificate(data=x509_path.read_bytes(), backend=default_backend())

        # common_name (optional, we can have only sans)
        common_name = x509_object.subject.get_attributes_for_oid(NameOID.COMMON_NAME)
        if len(common_name) > 0:
            self.common_name = common_name[0].value

        # sans
        try:
            for dns_name in x509_object.extensions.get_extension_for_oid(ExtensionOID.SUBJECT_ALTERNATIVE_NAME).value:
                self.sans.append(dns_name.value)
        except ExtensionNotFound:
            # no san in that certificate
            pass

        self.expiration = x509_object.not_valid_after
        return self

    def is_expired(self, expiration_delay=0):
        """Check for expiration

        :param expiration_delay: Number of days before real expiration we consider a renewal needed
        :type expiration_delay: int
        :return: True is certificate is going to expire in less than expiration_delay days
        :rtype: bool
        """
        status = 'valid'
        now = datetime.datetime.utcnow()
        if self.expiration < now:
            status = 'expired'
        else:
            status += ' (%s days remaining)' % (self.expiration.date() - now.date()).days
        logger.info("{0} - {1} => {2}".format(
            self.common_name, self.expiration.isoformat(), status))
        return (self.expiration.date() - now.date()).days <= expiration_delay

    def is_same(self, common_name=None, sans=None, exact_match=False):
        """Check if current certificate is covering all specified domains

        :param common_name: Common name
        :type common_name: str
        :param sans: list of Subject Alternate Names
        :type sans: list
        :param exact_match: if True, certificate must exactly match input domains
            if False, input domain will also match wilcard certificate and additional domains in certificate will
            be ignored
        :type exact_match: bool
        :return: True is certificate is already covering all domains
        """

        # build unique list of domains that is expected to be covered
        input_domains = get_domains(common_name=common_name, sans=sans)

        # check if some domains are missing
        return is_domain_list_matching(
            domains_to_check=input_domains,
            reference_domains=self.domains,
            exact_match=exact_match,
        )

    @property
    def domains(self):
        # build unique list of domains that is expected to be covered
        return get_domains(common_name=self.common_name, sans=self.sans)



import pytest

from autossl.server import base


@pytest.fixture(scope="module")
def base_server():
    # return local server instance
    yield base.Server(crt_name='example.autossl.com')


def test_get_description(base_server):
    with pytest.raises(NotImplementedError):
        base_server.get_description()


def test_deploy_cert(base_server):
    with pytest.raises(NotImplementedError):
        base_server.deploy_cert(key=None, cert=None, chain=None)


def test_create_acme_challenge(base_server):
    with pytest.raises(NotImplementedError):
        base_server.create_acme_challenge(token=None, key_authorization=None)


def test_delete_acme_challenge(base_server):
    with pytest.raises(NotImplementedError):
        base_server.delete_acme_challenge(token=None)


def test_get_certificate_information(base_server):
    with pytest.raises(NotImplementedError):
        base_server.get_certificate_information()


def test_is_same(base_server):
    with pytest.raises(NotImplementedError):
        base_server.is_same(common_name=None, sans=None, exact_match=False)


def test_is_expired(base_server):
    with pytest.raises(NotImplementedError):
        base_server.is_expired()



import json


class AutoSslException(Exception):
    """Generic exception for autossl

    Allow to chain exceptions keeping track of origin exception
    """
    def __init__(self, msg, original_exception=None):
        message = msg
        if original_exception:
            message += ": %s" % original_exception
        super(AutoSslException, self).__init__(message)
        self.__cause__ = original_exception
        self.__suppress_context__ = True


class HttpCodeException(AutoSslException):
    def __init__(self, request_exception):
        """Exception raised when received Http response has an invalid http code

        :param request_exception: requests HTTP exception
        :type request_exception: requests.exceptions.HTTPError
        """
        try:
            response_body_json = request_exception.response.json()
            response_body_text = json.dumps(response_body_json, indent=4, sort_keys=True)
        except ValueError:
            response_body_json = None
            response_body_text = request_exception.response.text
        exception_message = "HTTPError: => %s %s : %s" % (request_exception.request.method,
                                                          request_exception.request.url,
                                                          response_body_text)
        super(HttpCodeException, self).__init__(exception_message, original_exception=request_exception)
        self.status_code = request_exception.response.status_code
        self.response_body_text = response_body_text
        self.response_body_json = response_body_json


class NotFound(AutoSslException):
    """Requested data not found"""
    pass


class SslBlueprintInconsistency(AutoSslException):
    """SSL blueprint definition contains inconsistencies"""
    pass


class InvalidCertificate(AutoSslException):
    """Certificate is not matching expected criteria"""
    pass


class InvalidTrustChain(InvalidCertificate):
    """Certificate is not compatible with CA certificate specified"""
    pass


class KeyMismatch(InvalidCertificate):
    """Certificate does not match private key"""
    pass


class ExpiredCertificate(InvalidCertificate):
    """Certificate is expiring"""
    pass


class DefinitionMismatch(InvalidCertificate):
    """Certificate is not matching blueprint definition"""
    pass


class CertificateNotFound(NotFound):
    """Requested certificate not present on server"""
    pass


class DeployCertificateError(AutoSslException):
    """Unexpected error when trying to deploy new certificate"""
    pass



from __future__ import absolute_import, print_function

import argparse
import fnmatch
import logging
import os
from six.moves import configparser
import sys

import autossl
from autossl import manager, util

logging.basicConfig(level=logging.INFO)

logger = logging.getLogger('autossl')


def parse_arguments(args=None):
    parser = argparse.ArgumentParser(description=autossl.__description__)

    parser.add_argument('--credentials', action='append', default=[],
                        help='Generic user/password credentials. Format: name:user:password')

    parser.add_argument('--credentials-file',
                        type=util.Path,
                        default=util.Path('~/.autossl').expanduser(),
                        help='Local file to store credentials.')

    parser.add_argument('--debug', action="store_const", const=logging.DEBUG, default=logging.INFO,
                        help='Use DEBUG log level rather than INFO')

    parser.add_argument("--staging", action='store_true',
                        help="Testing mode (for example, use staging CA servers)")

    parser.add_argument("--config",
                        type=util.Path,
                        help="Path to a file containing common configuration. Same format than SSL blueprint.")

    parser.add_argument("--blueprint",
                        required=False,
                        type=util.Path,
                        help="Path to the definition of certificates. Can be single blueprint or a folder")

    action_subparser = parser.add_subparsers(title="action", dest="action")
    action_subparser.required = True  # needed for python3 compatibility

    ################
    # VERSION
    ################
    action_subparser.add_parser("version", help="Display autossl current version")

    ########################################
    # Parser to check if certificate needs to be renewed
    ########################################
    action_subparser.add_parser("check", help="Check for expired certificate.")

    ########################################
    # Parser to request certificate renewal
    ########################################
    action_parser_renew = action_subparser.add_parser("renew", help="Request certificate renewal")
    action_parser_renew.add_argument("--force",
                                     action='store_true',
                                     help="Renew certificate now (even if current one is still valid)")

    ########################################
    # Parser to deploy certificate on server
    ########################################
    action_parser_deploy = action_subparser.add_parser("deploy", help="Deploy certificate")
    action_parser_deploy.add_argument('-t', '--tracking-record',
                                      help='Tracking record ID for the change, '
                                           'and also potentially containing certificate to deploy.')
    action_parser_deploy.add_argument('-k', '--private-key',
                                      type=util.Path,
                                      help='Local path to certificate private key')
    action_parser_deploy.add_argument('-c', '--certificate',
                                      type=util.Path,
                                      help='Local path to ssl signed certificate')
    action_parser_deploy.add_argument('--all-servers', action='store_true', default=False,
                                      help='Deploy certificate on all servers rather than just the ones out of synch.')

    return parser.parse_args(args)


def main(args=None):
    if args is None:
        args = sys.argv[1:]

    options = parse_arguments(args=args)

    logger.setLevel(options.debug or logger.level)

    # this is the only action that does not require authentication
    if options.action == "version":
        print(display_version())
        return

    # actions that can be done on multiple blueprints
    if options.action in ['check', 'renew']:

        # build list of all blueprints to process
        blueprints = []
        if options.blueprint.is_file():
            blueprints.append(options.blueprint)
        elif options.blueprint.is_dir():
            for root_path, _, filenames in os.walk(options.blueprint):
                for filename in fnmatch.filter(filenames, '*.yaml'):
                    blueprints.append(util.Path(root_path) / filename)
        else:
            raise IOError("Invalid path specified: '%s'" % options.blueprint)

        # process blueprints 1 by 1
        for blueprint in blueprints:

            my_ssl_manager = manager.SslManager(
                global_config=options.config,
                blueprint_path=blueprint,
                credentials=parse_credentials(
                    cli_credentials=options.credentials,
                    credentials_file=options.credentials_file,
                ),
                staging=options.staging,
            )
            if options.action == 'check':
                logger.info("Processing blueprint %s" % blueprint)
                to_be_renewed, servers_to_update = my_ssl_manager.get_renewal_status()
                if not to_be_renewed and len(servers_to_update) == 0:
                    logger.info("Certificate and all servers up to date for '%s'. Nothing to do.",
                                my_ssl_manager.ssl_blueprint.name)
            elif options.action == 'renew':
                my_ssl_manager.renew(force=options.force)

    # action on a single blueprint
    elif options.action == 'deploy':

        my_ssl_manager = manager.SslManager(
            global_config=options.config,
            blueprint_path=options.blueprint,
            credentials=parse_credentials(
                cli_credentials=options.credentials,
                credentials_file=options.credentials_file
            ),
            staging=options.staging,
        )

        my_ssl_manager.deploy(
            tracking_record_id=options.tracking_record,
            certificate_path=options.certificate,
            private_key_path=options.private_key,
            all_servers=options.all_servers,
        )


def display_version():
    return "Autossl version {version}".format(version=autossl.__version__)


def parse_credentials(cli_credentials, credentials_file):
    """Perform first parsing of input credentials to extract credential name
    rest of the parsing will be done later depending on credential type

    :param cli_credentials: list of raw credentials from command line
    :type cli_credentials: list
    :param credentials_file: path to a local file containing credentials
    :type credentials_file: pathlib.Path or None
    :return: consolidated credentials information from credentials file and command line
    :rtype: dict
    """
    credentials_dict = dict()

    # read first credentials from local credentials file
    if credentials_file and credentials_file.exists():
        credentials_config = configparser.ConfigParser()
        credentials_config.read(str(credentials_file))

        for credential_name in credentials_config.sections():
            credential_parameters_dict = {}
            for credential_parameter in credentials_config.options(credential_name):
                credential_parameters_dict[credential_parameter] = \
                    credentials_config.get(credential_name, credential_parameter)
            # add retrieved credential to global credential dict
            credentials_dict[credential_name] = credential_parameters_dict

    # then override with any command line credential provided
    for cmd_line_credential in cli_credentials:
        # credential name and content is always separated by a ':'
        # other fields can be separated by any other separator specified in global credentials config (default to ':')
        credential_table = cmd_line_credential.split(':', 1)
        name = credential_table[0]
        raw_content = credential_table[1] if len(credential_table) > 1 else None
        credentials_dict[name] = {
            'raw_content': raw_content
        }

    return credentials_dict


if __name__ == '__main__':
    main()



from datetime import datetime

import pytest

from autossl import ssl, exception

from tests import util as tests_util


@pytest.mark.parametrize('cert1,cert2,is_same', [
    # fully identical
    ({'common_name': 'name1', 'sans': ['name2', 'name3'], 'expiration': datetime(2020, 2, 3, 14, 38, 21)},
     {'common_name': 'name1', 'sans': ['name2', 'name3'], 'expiration': datetime(2020, 2, 3, 14, 38, 21)},
     True),
    # fully identical but different san order
    ({'common_name': 'name1', 'sans': ['name2', 'name3'], 'expiration': datetime(2020, 2, 3, 14, 38, 21)},
     {'common_name': 'name1', 'sans': ['name3', 'name2'], 'expiration': datetime(2020, 2, 3, 14, 38, 21)},
     True),
    # fully common_name
    ({'common_name': 'name1', 'sans': ['name2', 'name3'], 'expiration': datetime(2020, 2, 3, 14, 38, 21)},
     {'common_name': 'name4', 'sans': ['name2', 'name3'], 'expiration': datetime(2020, 2, 3, 14, 38, 21)},
     False),
    # different san
    ({'common_name': 'name1', 'sans': ['name2', 'name3'], 'expiration': datetime(2020, 2, 3, 14, 38, 21)},
     {'common_name': 'name1', 'sans': ['name2', 'name4'], 'expiration': datetime(2020, 2, 3, 14, 38, 21)},
     False),
    # different expiration
    ({'common_name': 'name1', 'sans': ['name2', 'name3'], 'expiration': datetime(2020, 2, 3, 14, 38, 21)},
     {'common_name': 'name1', 'sans': ['name2', 'name3'], 'expiration': datetime(2021, 2, 3, 14, 38, 21)},
     False),
])
def test_ssl_blueprint___eq__(cert1, cert2, is_same):
    assert (ssl.SslCertificate(**cert1) == ssl.SslCertificate(**cert2)) is is_same


def test_missing_blueprint():
    with pytest.raises(IOError):
        ssl.SslBlueprint('dummy/path')


def test_ssl_blueprint_no_server():
    ssl_blueprint = ssl.SslBlueprint(tests_util.DATA_PATH / 'tst.ov.example.com_no-server.yaml')
    assert ssl_blueprint.name == 'auto_tst.autossl.example.com'
    assert ssl_blueprint.certificate.certificate_type == 'OV'
    assert ssl_blueprint.certificate.certificate_authority == 'Sectigo'
    assert ssl_blueprint.certificate.common_name == 'tst.autossl.example.com'
    assert len(ssl_blueprint.servers) == 0


def test_ov_ssl_blueprint():
    ssl_blueprint = ssl.SslBlueprint(tests_util.DATA_PATH / 'tst.ov.example.com.yaml')
    assert ssl_blueprint.name == 'auto_tst.autossl.example.com'
    assert ssl_blueprint.certificate.certificate_type == 'OV'
    assert ssl_blueprint.certificate.certificate_authority == 'Sectigo'
    assert ssl_blueprint.certificate.common_name == 'tst.autossl.example.com'
    assert ssl_blueprint.certificate.renewal_delay == 30
    assert len(ssl_blueprint.servers) == 1
    assert len(ssl_blueprint.certificate.sans) == 5
    assert ssl_blueprint.certificate.organization['company_name'] == 'Autossl corporation'
    assert ssl_blueprint.certificate.organization['street_address'] == 'Newbury street'
    assert ssl_blueprint.certificate.organization['city'] == 'Boston'
    assert ssl_blueprint.certificate.organization['state'] == 'Massachusetts'
    assert ssl_blueprint.certificate.organization['postal_code'] == '02115'
    assert ssl_blueprint.certificate.organization['country_code'] == 'US'


def test_dv_ssl_blueprint():
    ssl_blueprint = ssl.SslBlueprint(tests_util.DATA_PATH / 'tst.dv.example.com.yaml')
    assert ssl_blueprint.name == 'auto_tst.autossl.example.com'
    assert ssl_blueprint.certificate.certificate_type == 'DV'
    assert ssl_blueprint.certificate.certificate_authority == 'LetsEncrypt'
    assert ssl_blueprint.certificate.common_name == 'tst.autossl.example.com'
    assert ssl_blueprint.certificate.renewal_delay == 30
    assert len(ssl_blueprint.servers) == 2
    assert len(ssl_blueprint.certificate.sans) == 5
    assert ssl_blueprint.certificate.organization is None


def test_ssl_blueprint_with_global_config():
    ssl_blueprint = ssl.SslBlueprint(
        yaml_path=tests_util.DATA_PATH / 'tst.ov.example.com_minimal.yaml',
        global_config_path=tests_util.DATA_PATH / 'global_config.yaml',
    )
    assert ssl_blueprint.name == 'auto_tst.autossl.example.com'
    assert ssl_blueprint.certificate.certificate_type == 'DV'
    assert ssl_blueprint.certificate.certificate_authority == 'LetsEncrypt'
    assert ssl_blueprint.certificate.common_name == 'tst.autossl.example.com'
    assert ssl_blueprint.certificate.renewal_delay == 30
    assert len(ssl_blueprint.servers) == 1
    assert len(ssl_blueprint.certificate.sans) == 5
    assert ssl_blueprint.certificate.organization['company_name'] == 'Autossl corporation'
    assert ssl_blueprint.certificate.organization['street_address'] == 'Newbury street'
    assert ssl_blueprint.certificate.organization['city'] == 'Boston'
    assert ssl_blueprint.certificate.organization['state'] == 'Massachusetts'
    assert ssl_blueprint.certificate.organization['postal_code'] == '02115'
    assert ssl_blueprint.certificate.organization['country_code'] == 'US'


def test_ssl_blueprint_no_common_name(tmp_path):
    blueprint_content = u"""
---
name: auto_tst.autossl.example.com

servers:
  - type: autossl.server.local.LocalServer
    parameters:
        path: /etc/ssl/my_certificates

certificate:
  type: DV
  certificate_authority: LetsEncrypt
  san:
    - tst.autossl.example.com
    - uat.tst.autossl.example.com
    - pit.tst.autossl.example.com
...
    """
    blueprint_path = tmp_path / 'blueprint.yaml'
    blueprint_path.write_text(blueprint_content, encoding='utf-8')

    ssl_blueprint = ssl.SslBlueprint(str(blueprint_path))
    assert ssl_blueprint.certificate.common_name is None
    assert len(ssl_blueprint.certificate.sans) == 3


def test_ssl_blueprint_no_san(tmp_path):
    blueprint_content = u"""
---
name: auto_tst.autossl.example.com

servers:
  - type: autossl.server.local.LocalServer
    parameters:
        path: /etc/ssl/my_certificates

certificate:
  type: DV
  certificate_authority: LetsEncrypt
  common_name: tst.autossl.example.com
...
        """
    blueprint_path = tmp_path / 'blueprint.yaml'
    blueprint_path.write_text(blueprint_content, encoding='utf-8')

    ssl_blueprint = ssl.SslBlueprint(str(blueprint_path))
    assert ssl_blueprint.certificate.common_name == 'tst.autossl.example.com'
    assert len(ssl_blueprint.certificate.sans) == 0


def test_ssl_blueprint_no_commmon_name_no_san(tmp_path):
    blueprint_content = u"""
---
name: auto_tst.autossl.example.com

servers:
  - type: autossl.server.local.LocalServer
    parameters:
        path: /etc/ssl/my_certificates

certificate:
  type: DV
  certificate_authority: LetsEncrypt
...
        """
    blueprint_path = tmp_path / 'blueprint.yaml'
    blueprint_path.write_text(blueprint_content, encoding='utf-8')

    with pytest.raises(ValueError):
        ssl.SslBlueprint(str(blueprint_path))


@pytest.mark.parametrize('common_name,is_valid', [
    ('test2_valid-test.example.com', True),
    ('*.example.com', True),
    (' test.example.com', False),
    ('test.example.com ', False),
    ('test.*.com', False),
    ('%1.example.com', False),
])
def test_ssl_blueprint_validate_common_name(tmp_path, common_name, is_valid):
    blueprint_content = u"""
---
name: auto_tst.autossl.example.com

servers:
  - type: autossl.server.local.LocalServer
    parameters:
        path: /etc/ssl/my_certificates

certificate:
  type: DV
  certificate_authority: LetsEncrypt
  common_name: '{}'
...
        """.format(common_name)
    blueprint_path = tmp_path / 'blueprint.yaml'
    blueprint_path.write_text(blueprint_content, encoding='utf-8')

    if is_valid:
        ssl.SslBlueprint(str(blueprint_path))
    else:
        with pytest.raises(ValueError):
            ssl.SslBlueprint(str(blueprint_path))


def test_get_domains():
    ssl_blueprint = ssl.SslBlueprint(tests_util.DATA_PATH / 'tst.dv.example.com.yaml')
    assert ssl_blueprint.domains == {
        'tst.autossl.example.com',
        'uat.tst.autossl.example.com',
        'pit.tst.autossl.example.com',
        'cit.tst.autossl.example.com',
        'mgt.tst.autossl.example.com'
    }


def test_is_domain_matching():
    assert ssl.is_domain_matching('test.example.com', 'test.example.com')
    assert ssl.is_domain_matching('test.example.com', 'test.example.com', True)
    assert ssl.is_domain_matching('test.example.com', 'test.example.com', False)

    assert ssl.is_domain_matching('test.example.com', 'test2.example.com') is False
    assert ssl.is_domain_matching('test.example.com', 'test2.example.com', True) is False
    assert ssl.is_domain_matching('test.example.com', 'test2.example.com', False) is False

    assert ssl.is_domain_matching('test.example.com', '*.example.com') is True
    assert ssl.is_domain_matching('test.example.com', '*.example.com', True) is False
    assert ssl.is_domain_matching('test.example.com', '*.example.com', False) is True


def test_is_domain_list_matching():
    assert ssl.is_domain_list_matching(['test.example.com'], ['test.example.com'])
    assert ssl.is_domain_list_matching(['test.example.com'], ['test.example.com', 'test2.example.com'])
    assert ssl.is_domain_list_matching(['test.example.com', 'test2.example.com'], ['test.example.com']) is False
    assert ssl.is_domain_list_matching(['test.example.com', 'test2.example.com'], ['*.example.com'])
    assert ssl.is_domain_list_matching(
        ['test.example.com', 'test2.example.com'], ['*.example.com'], exact_match=True) is False


def test_get_config():
    ssl_blueprint = ssl.SslBlueprint(tests_util.DATA_PATH / 'tst.dv.example.com.yaml')

    assert ssl_blueprint.get_config(name='tracking', path=['dummy_path'], default=[]) == []
    assert ssl_blueprint.get_config(name='tracking', path=None, default=None) is None
    assert ssl_blueprint.get_config(name='storage', path=None, default=None) == {
        'credentials': 'credential_1',
        'data': [{'type': 'key'}, {'type': 'csr'}, {'type': 'crt'}],
        'parameters': {
            'git_url': 'https://git.autossl.com/git/scm/ssl/certificates.git',
            'config_user_name': 'Test User',
            'config_user_email': 'test.user@autossl.com',
        },
        'type': 'autossl.storage.gitscm.GitStorage'}


def test_check_chain_of_trust(tmp_path):
    crt_path = tmp_path / 'local.crt'
    ca_crt_path = tmp_path / 'local_ca.crt'
    ca_key_path = tmp_path / 'local_ca.key'
    # generate CA certificate
    key, crt = tests_util.create_ca_certificate(ca_name='Autossl')
    ca_crt_path.write_bytes(crt)
    ca_key_path.write_bytes(key)

    # sign a new certificate with the CA
    _, csr_path = ssl.generate_csr(name='autossl_cert', common_name='test.autossl.com', output_path=str(tmp_path))
    crt_content = tests_util.create_signed_certificate(
        csr_path=csr_path,
        ca_crt_path=ca_crt_path,
        ca_key_path=ca_key_path,
    )
    crt_path.write_bytes(crt_content)

    # valid trust chain should no raise any error
    ssl.check_chain_of_trust(
        chain_of_trust=[crt.decode('utf-8')],  # Chain of trust comes normally from SSL blueprint so it not in bytes
        crt_path=crt_path,
    )

    # generate self-signed certificate
    self_signed_key_path, self_signed_crt_path = tests_util.create_self_signed_certificate(
        crt_name="self_signed_local.crt",
        output_path=tmp_path,
        common_name='self_signed.test.autossl.com',
    )

    # self signed certificate should not be validated by this CA
    with pytest.raises(exception.InvalidTrustChain):
        ssl.check_chain_of_trust(
            chain_of_trust=[crt.decode('utf-8')],  # Chain of trust comes normally from SSL blueprint so it not in bytes
            crt_path=self_signed_crt_path,
        )



"""
Tests for Incapsula server implementation
"""
import base64
import collections
import datetime
import json

import mock
import pytest


from autossl import exception, ssl, util
from autossl.server import incapsula

from tests import util as tests_util

DATA_PATH = util.Path(__file__).parent / 'data'


CertificateKeyPair = collections.namedtuple('CertificateKeyPair', 'key crt')
SITE_ID = '12345678'
API_KEY = '6f4d1878-f2c6-446b-8932-636b8f1705a7'
API_ID = '12345'
CHAIN_OF_TRUST = [
    """"-----BEGIN CERTIFICATE-----
MIIEkjCCA3qgAwIBAgIQCgFBQgAAAVOFc2oLheynCDANBgkqhkiG9w0BAQsFADA/MSQwIgYDVQQK
ExtEaWdpdGFsIFNpZ25hdHVyZSBUcnVzdCBDby4xFzAVBgNVBAMTDkRTVCBSb290IENBIFgzMB4X
DTE2MDMxNzE2NDA0NloXDTIxMDMxNzE2NDA0NlowSjELMAkGA1UEBhMCVVMxFjAUBgNVBAoTDUxl
dCdzIEVuY3J5cHQxIzAhBgNVBAMTGkxldCdzIEVuY3J5cHQgQXV0aG9yaXR5IFgzMIIBIjANBgkq
hkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEAnNMM8FrlLke3cl03g7NoYzDq1zUmGSXhvb418XCSL7e4
S0EFq6meNQhY7LEqxGiHC6PjdeTm86dicbp5gWAf15Gan/PQeGdxyGkOlZHP/uaZ6WA8SMx+yk13
EiSdRxta67nsHjcAHJyse6cF6s5K671B5TaYucv9bTyWaN8jKkKQDIZ0Z8h/pZq4UmEUEz9l6YKH
y9v6Dlb2honzhT+Xhq+w3Brvaw2VFn3EK6BlspkENnWAa6xK8xuQSXgvopZPKiAlKQTGdMDQMc2P
MTiVFrqoM7hD8bEfwzB/onkxEz0tNvjj/PIzark5McWvxI0NHWQWM6r6hCm21AvA2H3DkwIDAQAB
o4IBfTCCAXkwEgYDVR0TAQH/BAgwBgEB/wIBADAOBgNVHQ8BAf8EBAMCAYYwfwYIKwYBBQUHAQEE
czBxMDIGCCsGAQUFBzABhiZodHRwOi8vaXNyZy50cnVzdGlkLm9jc3AuaWRlbnRydXN0LmNvbTA7
BggrBgEFBQcwAoYvaHR0cDovL2FwcHMuaWRlbnRydXN0LmNvbS9yb290cy9kc3Ryb290Y2F4My5w
N2MwHwYDVR0jBBgwFoAUxKexpHsscfrb4UuQdf/EFWCFiRAwVAYDVR0gBE0wSzAIBgZngQwBAgEw
PwYLKwYBBAGC3xMBAQEwMDAuBggrBgEFBQcCARYiaHR0cDovL2Nwcy5yb290LXgxLmxldHNlbmNy
eXB0Lm9yZzA8BgNVHR8ENTAzMDGgL6AthitodHRwOi8vY3JsLmlkZW50cnVzdC5jb20vRFNUUk9P
VENBWDNDUkwuY3JsMB0GA1UdDgQWBBSoSmpjBH3duubRObemRWXv86jsoTANBgkqhkiG9w0BAQsF
AAOCAQEA3TPXEfNjWDjdGBX7CVW+dla5cEilaUcne8IkCJLxWh9KEik3JHRRHGJouM2VcGfl96S8
TihRzZvoroed6ti6WqEBmtzw3Wodatg+VyOeph4EYpr/1wXKtx8/wApIvJSwtmVi4MFU5aMqrSDE
6ea73Mj2tcMyo5jMd6jmeWUHK8so/joWUoHOUgwuX4Po1QYz+3dszkDqMp4fklxBwXRsW10KXzPM
TZ+sOPAveyxindmjkW8lGy+QsRlGPfZ+G6Z6h7mjem0Y+iWlkYcV4PIWL1iwBi8saCbGS5jN2p8M
+X+Q7UNKEkROb3N6KOqkqm57TH2H3eDJAkSnh6/DNFu0Qg==
-----END CERTIFICATE-----
""",
    """"-----BEGIN CERTIFICATE-----
MIIDSjCCAjKgAwIBAgIQRK+wgNajJ7qJMDmGLvhAazANBgkqhkiG9w0BAQUFADA/MSQwIgYDVQQK
ExtEaWdpdGFsIFNpZ25hdHVyZSBUcnVzdCBDby4xFzAVBgNVBAMTDkRTVCBSb290IENBIFgzMB4X
DTAwMDkzMDIxMTIxOVoXDTIxMDkzMDE0MDExNVowPzEkMCIGA1UEChMbRGlnaXRhbCBTaWduYXR1
cmUgVHJ1c3QgQ28uMRcwFQYDVQQDEw5EU1QgUm9vdCBDQSBYMzCCASIwDQYJKoZIhvcNAQEBBQAD
ggEPADCCAQoCggEBAN+v6ZdQCINXtMxiZfaQguzH0yxrMMpb7NnDfcdAwRgUi+DoM3ZJKuM/IUmT
rE4Orz5Iy2Xu/NMhD2XSKtkyj4zl93ewEnu1lcCJo6m67XMuegwGMoOifooUMM0RoOEqOLl5CjH9
UL2AZd+3UWODyOKIYepLYYHsUmu5ouJLGiifSKOeDNoJjj4XLh7dIN9bxiqKqy69cK3FCxolkHRy
xXtqqzTWMIn/5WgTe1QLyNau7Fqckh49ZLOMxt+/yUFw7BZy1SbsOFU5Q9D8/RhcQPGX69Wam40d
utolucbY38EVAjqr2m7xPi71XAicPNaDaeQQmxkqtilX4+U9m5/wAl0CAwEAAaNCMEAwDwYDVR0T
AQH/BAUwAwEB/zAOBgNVHQ8BAf8EBAMCAQYwHQYDVR0OBBYEFMSnsaR7LHH62+FLkHX/xBVghYkQ
MA0GCSqGSIb3DQEBBQUAA4IBAQCjGiybFwBcqR7uKGY3Or+Dxz9LwwmglSBd49lZRNI+DT69ikug
dB/OEIKcdBodfpga3csTS7MgROSR6cz8faXbauX+5v3gTt23ADq1cEmv8uXrAvHRAosZy5Q6XkjE
GB5YGV8eAlrwDPGxrancWYaLbumR9YbK+rlmM6pZW87ipxZzR8srzJmwN0jP41ZL9c8PDHIyh8bw
RLtTcm1D9SZImlJnt1ir/md2cXjbDaJWFBM5JDGFoqgCWjBH4d1QB7wCCZAA62RjYJsWvIjJEubS
fZGL+T0yjWW06XyxV3bqxbYoOb8VZRzI9neWagqNdwvYkQsEjgfbKbYK7p2CNTUQ
-----END CERTIFICATE-----
"""
]


class MockResponse:
    """"Mock requests.Session response object"""

    def __init__(self, content_path):
        self.status_code = 200
        self.content = content_path.read_text()

    def json(self):
        return json.loads(self.content)


@pytest.fixture(scope="function")
def incapsula_json_response_first_call():
    return 'incapsula_get_site_ok.json'


@pytest.fixture(scope="function")
def incapsula_json_response_second_call():
    return 'incapsula_get_site_ok.json'


# use function scope to make sure we can check only the mock calls of current scenario
@pytest.fixture(scope="function")
def incapsula_site(incapsula_json_response_first_call, incapsula_json_response_second_call):
    # patched_session will have a list attribute mock_calls that will contain (after __init__) at least the following:
    # mock_calls[0] is Session object creation
    # mock_calls[1] is the call to Session.mount() in order to enable http retries
    # mock_calls[2] is the retrieval of Incapsula site information
    with mock.patch('autossl.server.incapsula.requests.Session') as patched_session:
        # return local server instance
        patched_session.return_value = mock.MagicMock(post=mock.MagicMock(
                side_effect=[
                    # first call to retrieve site config is always ok
                    MockResponse(content_path=DATA_PATH / incapsula_json_response_first_call),
                    # second call can return either ok or an error
                    MockResponse(content_path=DATA_PATH / incapsula_json_response_second_call),
                    # we can have extra calls in case of retries, let's consider retry works
                    MockResponse(content_path=DATA_PATH / 'incapsula_get_site_ok.json'),
                ]
            )
        )
        incapsula_server = incapsula.IncapsulaSite(API_KEY, API_ID, site_id=SITE_ID, crt_name='tst.example.autossl.com')

        yield incapsula_server, patched_session


@pytest.fixture(scope="function")
def keypair(tmp_path):
    """Generate SSL keypair: private key + certificate"""
    common_name = 'tst.autossl.example.com'
    crt_name = tests_util.get_random_ascii_string()
    key_path = tmp_path / (crt_name + '.key')
    key_content, csr_path = ssl.generate_csr(name=crt_name, common_name=common_name, output_path=tmp_path)
    crt_path = tmp_path / (crt_name + '.crt')
    crt_content = tests_util.create_signed_certificate(
        csr_path=csr_path,
        certificate_validity_days=10,
    )
    crt_path.write_bytes(crt_content)
    key_path.write_bytes(key_content)

    yield CertificateKeyPair(key_path, crt_path)


def test_init(incapsula_site):
    # ensure incapsula_server api is well initialized
    assert incapsula_site[0].api_key == API_KEY
    assert incapsula_site[0].api_id == API_ID
    assert incapsula_site[0].site_id == SITE_ID
    assert incapsula_site[0].basic_params == {'api_key': API_KEY, 'api_id': API_ID, 'site_id': SITE_ID}
    assert isinstance(incapsula_site[0].site_status, dict)

    # ensure rest query to incapsula is well formed
    # mock_calls[0] is Session object creation
    # mock_calls[1] is the call to Session.mount() in order to enable http retries
    assert incapsula_site[1].mock_calls[2] == mock.call().post(
        data={
            'api_key': API_KEY,
            'site_id': SITE_ID,
            'api_id': API_ID,
        },
        url='https://my.incapsula.com:443/api/prov/v1/sites/status'
    )


def test_get_description(incapsula_site):
    assert incapsula_site[0].get_description() == '[IncapsulaSite - tst.example.autossl.com (12345678)]'


@pytest.mark.parametrize('incapsula_json_response_second_call', [
    'incapsula_get_site_ok.json',
    'incapsula_get_site_error.json'
])
def test_deploy_cert(incapsula_site, keypair):
    # deploy key + cert
    incapsula_site[0].deploy_cert(key=keypair.key, cert=keypair.crt)

    # ensure rest query to incapsula is well formed
    assert incapsula_site[1].mock_calls[3] == mock.call().post(
        data={
            'api_key': '6f4d1878-f2c6-446b-8932-636b8f1705a7',
            'site_id': '12345678',
            'api_id': '12345',
            'certificate': base64.b64encode(keypair.crt.read_bytes()),
            'private_key': base64.b64encode(keypair.key.read_bytes()),
        },
        url='https://my.incapsula.com:443/api/prov/v1/sites/customCertificate/upload'
    )


def test_get_certificate_information(incapsula_site):
    certificate_information = incapsula_site[0].get_certificate_information()

    assert certificate_information.common_name == 'tst.example.autossl.com'
    assert certificate_information.sans == []
    assert certificate_information.expiration == datetime.datetime(
        year=2019, month=3, day=17, hour=9, minute=17, second=43)


# customize incaspula deploy to return an error
@pytest.mark.parametrize('incapsula_json_response_first_call', ['incapsula_get_site_no_ssl_certificate.json'])
def test_get_certificate_information_no_custom_ssl_certificate(incapsula_site):
    with pytest.raises(exception.CertificateNotFound):
        incapsula_site[0].get_certificate_information()


def test_is_same(incapsula_site):
    assert incapsula_site[0].is_same(common_name='tst.example.autossl.com', sans=None) is True
    assert incapsula_site[0].is_same(common_name=None, sans=['tst.example.autossl.com']) is True
    assert incapsula_site[0].is_same(common_name=None, sans=['test1.tst.example.autossl.com',
                                                             'test2.tst.example.autossl.com',
                                                             'tst.example.autossl.com']) is True
    assert incapsula_site[0].is_same(common_name='*.example.autossl.com', sans=None) is True
    assert incapsula_site[0].is_same(common_name='test2.tst.example.autossl.com', sans=None) is False
    assert incapsula_site[0].is_same(common_name=None, sans=['test2.tst.example.autossl.com',
                                                             'test-3.tst.example.autossl.com']) is False


def test_create_acme_challenge(incapsula_site):
    # call should do nothing and should not fail
    incapsula_site[0].create_acme_challenge(token=None, key_authorization=None)

    # ensure no external call has been made
    assert len(incapsula_site[1].mock_calls) == 3


def test_delete_acme_challenge(incapsula_site):
    # call should do nothing and should not fail
    incapsula_site[0].delete_acme_challenge(token=None)

    # ensure no external call has been made
    assert len(incapsula_site[1].mock_calls) == 3



"""
Tests for Local file logging tracking implementation
"""
import os
import pytest

from autossl import exception, ssl, util
from autossl.tracking import local, base

from tests import util as tests_util


@pytest.fixture(scope="module")
def local_logging_tracking():
    ssl_blueprint_name = 'tst.ov.example.com.yaml'
    ssl_blueprint_path = tests_util.DATA_PATH / ssl_blueprint_name

    with util.TempDir() as temp_folder:
        yield local.LocalFileTracking(
            ssl_blueprint_path=ssl_blueprint_path,
            log_folder=str(temp_folder.path),
        )


def test_create_basic(local_logging_tracking):
    # log file does not exist before create
    assert not os.path.exists(local_logging_tracking.log_file_path)
    local_logging_tracking.create(tracking_type=base.TrackingType.Renewal)
    # and is well created after call to create
    assert os.path.isfile(local_logging_tracking.log_file_path)

    assert local_logging_tracking.retrieve_data(name='log').startswith(
        b"Creation of the following certificate requested")


def test_save_data(local_logging_tracking):
    local_logging_tracking.save_data(name='data name', data_type=None, local_path=None, content=b'my data content')


def test_update(local_logging_tracking):
    local_logging_tracking.update(message='insert message')
    assert local_logging_tracking.retrieve_data(name='log') == b'insert message'


def test_refresh(local_logging_tracking):
    local_logging_tracking.refresh(record_id=None)


def test_retrieve_data(local_logging_tracking):
    with pytest.raises(ValueError):
        local_logging_tracking.retrieve_data()

    with pytest.raises(exception.NotFound):
        local_logging_tracking.retrieve_data(name='data_identifier')

    with pytest.raises(exception.NotFound):
        local_logging_tracking.retrieve_data(data_type=ssl.DataType.Certificate)

    local_logging_tracking.save_data(name='data_identifier', data_type=ssl.DataType.Certificate,
                                     content=b'my data content')
    assert local_logging_tracking.retrieve_data(name='data_identifier') == b'my data content'
    assert local_logging_tracking.retrieve_data(data_type=ssl.DataType.Certificate) == b'my data content'


def test_close_for_failure(local_logging_tracking):
    local_logging_tracking.close_for_failure(message='Operation failed !')
    assert local_logging_tracking.retrieve_data(name='log') == b'Operation failed !'


def test_close_for_success(local_logging_tracking):
    local_logging_tracking.close_for_success(message='Operation completed successfully.')
    assert local_logging_tracking.retrieve_data(name='log') == b'Operation completed successfully.'



"""
Script to check and renew automatically SSL certificates on a server
"""
# standard packages
import logging
import os
import shutil
import tempfile
import traceback

# external packages
from six.moves import input
import yaml

# local packages
from . import exception, credential, ssl, util, server, tracking

# list of available apis types
DEFAULT_APIS = ['storage', 'tracking']

# import packages embedded in base module
# external plugins will be imported later and only when/if needed

logger = logging.getLogger(__name__)


class SslManager(object):

    def __init__(self,
                 global_config=None,
                 blueprint_path=None,
                 credentials=None,
                 staging=True):
        """

        :param global_config: local path to global configuration file
        :type global_config: str or None
        :param blueprint_path: SSL blueprint path
        :type blueprint_path: str
        :param staging: Testing mode. Use staging or test server (when available).
        :type staging: bool
        """
        self.ssl_blueprint = ssl.SslBlueprint(global_config_path=global_config, yaml_path=blueprint_path) \
            if blueprint_path or global_config else None
        self.staging = staging
        self.credentials = credentials or {}

        # will be initialized at first use
        self._tracking_api = None
        self._storage_api = None
        self._acme_storage_api = None
        self._ca_manager_api = None

    def __get_credentials_parameters(self, name, parameters):
        self.credentials[name] = credential.get_credentials(
            name=name,
            global_config=self.ssl_blueprint.credentials,
            credentials=self.credentials,
            extra_parameters=parameters,
        )
        return self.credentials[name]

    def get_tracking_api(self):
        return self.__get_api(name='tracking')

    def get_storage_api(self):
        return self.__get_api(name='storage')

    def get_ca_storage_api(self):
        ca_storage_config = None
        # preliminary checks
        if self.ssl_blueprint is not None:
            ca_storage_config = self.ssl_blueprint.ca_config.get_storage_config()
        if ca_storage_config and ca_storage_config.get('type') is not None:
            return self.__get_api(name='ca_storage', api_config=ca_storage_config)
        # by default, use same storage for CA than default storage api
        return self.get_storage_api()

    def get_ca_manager_api(self):
        if self.ssl_blueprint is None:
            raise RuntimeError('No SSL blueprint specified, cannot instantiate CA manager api.')
        ca_config = self.ssl_blueprint.ca_config
        parameters = {
            'parameters': {
                'ca_config': ca_config,
                'storage_api': self.get_ca_storage_api(),
            }
        }
        if ca_config is not None:
            parameters['type'] = ca_config.ca_config.get('type') or 'autossl.ca_manager.base.CaManager'
            if 'parameters' in ca_config.ca_config:
                parameters['parameters'].update(ca_config.ca_config['parameters'])  # add parameters from blueprint
        return self.__get_api(name='ca_manager', api_config=parameters)

    def __get_api(self, name, api_config=None):
        """ dynamically instantiate any api with required parameters

        :param name:
        :return:
        """
        # preliminary checks
        if self.ssl_blueprint is None:
            raise RuntimeError('No SSL blueprint specified, cannot instantiate {} api.'.format(name))
        self_api_name = '_{}_api'.format(name)

        # dynamically set api if not already done
        if not hasattr(self, self_api_name):
            setattr(self, self_api_name, None)

        self_api_config = api_config or getattr(self.ssl_blueprint, name)
        # check first if api has already been instantiated
        if getattr(self, self_api_name) is None:
            dynamic_parameters = dict()
            if self_api_config is not None:
                self_api_config_type = self_api_config['type']
                # add dynamic credentials parameters
                if 'credentials' in self_api_config:
                    dynamic_parameters.update(
                        self.__get_credentials_parameters(
                            name=self_api_config['credentials'],
                            parameters=self_api_config.get('credentials_parameters')
                        )
                    )

                # then update with api server specific parameters from ssl blueprint
                dynamic_parameters.update(self_api_config.get('parameters', {}))
            else:
                # just return base api
                logger.warning("Use of base api as none explicitly specified for %s", name)
                self_api_config_type = 'autossl.{}.base.{}'.format(name, name.title())

            # allow staging mode to be set per api rather than only globally, using api parameters from config file.
            # Default remains global staging flag
            if 'staging' not in dynamic_parameters:
                dynamic_parameters['staging'] = self.staging

            # assigned instantiated api to current instance to reuse it for later calls
            class_ = util.str_to_class(self_api_config_type)
            setattr(self,
                    self_api_name,
                    class_(
                         ssl_blueprint_path=self.ssl_blueprint.consolidated_blueprint_path,
                         **dynamic_parameters)
                    )
        return getattr(self, self_api_name)

    def get_server_api(self, server_parameters):
        # preliminary checks
        if self.ssl_blueprint is None:
            raise RuntimeError('No SSL blueprint specified, cannot instantiate server api.')

        # dynamically instantiate server api with required parameters
        dynamic_parameters = {
            'crt_name': self.ssl_blueprint.name,
            'host': server_parameters.get('host'),
        }

        # add dynamic credentials parameters
        if 'credentials' in server_parameters:
            dynamic_parameters.update(self.__get_credentials_parameters(
                name=server_parameters['credentials'],
                parameters=server_parameters.get('credentials_parameters'),
            ))

        # then update with server parameters
        dynamic_parameters.update(server_parameters.get('parameters', {}))

        # allow staging mode to be set per server rather than only globally, using api parameters from config file.
        # Default remains global staging flag
        if 'staging' not in dynamic_parameters:
            dynamic_parameters['staging'] = self.staging

        class_ = util.str_to_class(server_parameters['type'])
        return class_(
            **dynamic_parameters
        )

    def get_renewal_status(self):
        """Get details status of the certificate for each server from blueprint: expired, modified, missing, ...

        :return: a 2-tuple with (Boolean renewal needed, Array servers to update)
        :rtype: tuple

        The checks performed are the following
         1) it is a new certificate
         2) cert is close to expiration
         3) cert definition has been modified (ex: new san)
         4) new server has been added
        """
        to_be_renewed = False
        servers_to_update = []

        with util.TempDir() as temp_folder:
            try:
                stored_certificate = self.get_certificate_information(working_directory=temp_folder.path)

                if stored_certificate.is_expired(expiration_delay=self.ssl_blueprint.certificate.renewal_delay):
                    to_be_renewed = True
                    logger.info("Certificate expiring soon for '{}'".format(self.ssl_blueprint.name))
                if not stored_certificate.is_same(common_name=self.ssl_blueprint.certificate.common_name,
                                                  sans=self.ssl_blueprint.certificate.sans,
                                                  exact_match=self.ssl_blueprint.certificate.exact_match):
                    to_be_renewed = True
                    logger.info("Certificate definition changed for '{}'".format(self.ssl_blueprint.name))

                # stored certificate does not match blueprint, need renewal in any case
                if to_be_renewed:
                    servers_to_update = self.ssl_blueprint.servers

                # starting here, we know we have a valid certificate in storage
                # we will now check if each server specified in blueprint is up to date
                else:
                    # loop on all servers to check if there is any action to perform
                    for server_config in self.ssl_blueprint.servers:
                        # select proper server api based on server type
                        # and retrieve certificate status directly from servers
                        try:
                            server_api = self.get_server_api(server_parameters=server_config)
                            server_cert = server_api.get_certificate_information()

                            # certificate definition changed
                            # or we have a more recent stored certificate to deploy
                            if not server_api.is_same(
                                    common_name=stored_certificate.common_name,
                                    sans=stored_certificate.sans,
                                    exact_match=self.ssl_blueprint.certificate.exact_match,
                            ) or stored_certificate.expiration > server_cert.expiration:
                                logger.info(
                                    "Stored certificate and server certificate are different for '{}'"
                                    " on server '{}'".format(self.ssl_blueprint.name, server_api.get_description()))
                                servers_to_update.append(server_config)

                        except exception.NotFound:
                            servers_to_update.append(server_config)

            except exception.NotFound:
                to_be_renewed = True
                servers_to_update = self.ssl_blueprint.servers
                logger.info("Certificate '{}' missing".format(self.ssl_blueprint.name))

        return to_be_renewed, servers_to_update

    def renew(self, force=False):
        """Request a renewal and proceed with automated renewal right after (if applicable)

        :param force: request renewal even if not needed
        :type force: bool
        """
        if self.request_renewal(force=force) is True:
            self.renew_certificate()

    def request_renewal(self, force=False):
        """Request renewal of the certificate for specified blueprint

        it is first checking that a renewal is needed.
        Then it is generating a new CSR for the specified blueprint.
        A new tracking record is created with CSR and blueprint attached
        If automated renewal is supported, certificate is generated automatically with CA and attached to TR
        Else, TR is simply sent to 'SSL Certificate Service' team

        :param force: request renewal even if not needed
        :type force: bool
        :return: True if a renewal is needed
        :rtype: bool
        """
        logger.info("Processing blueprint {}".format(self.ssl_blueprint.blueprint_path))
        to_be_renewed = False
        servers_to_update = []

        if force:
            logger.info("Force renewal for '{}'".format(self.ssl_blueprint.name))
            to_be_renewed = True
            servers_to_update = self.ssl_blueprint.servers
        else:
            to_be_renewed, servers_to_update = self.get_renewal_status()
            if not to_be_renewed and len(servers_to_update) == 0:
                logger.info("Certificate and all servers up to date for '{}'. Nothing to do.".format(
                    self.ssl_blueprint.name))
                return False

        if not force:
            if input('Continue ? (y/n)').lower() not in ['y', 'yes']:
                logger.info("Skip renewal...")
                return False

        # no valid certificate, request a new one
        if to_be_renewed:
            logger.info("Start renewal process for certificate '{}'".format(self.ssl_blueprint.name))

            content_name = self.ssl_blueprint.name
            with util.TempDir() as temp_folder:
                key_content = None
                # if reuse of private key is enabled
                # check if it is already existing for this certificate and retrieve it
                if self.ssl_blueprint.certificate.private_key_reuse is True:
                    try:
                        key_content = self.get_storage_api().retrieve_data(
                            name='{}.{}'.format(content_name, ssl.DataType.PrivateKey.value),
                            data_type=ssl.DataType.PrivateKey,
                        )
                    except exception.NotFound:
                        # else a new one will be created automatically
                        pass

                # extract optional organization parameters
                organization_parameters = self.ssl_blueprint.certificate.organization or {}

                # create CSR (and optionally new private key if not retrieved from secret server)
                key_content, csr_path = ssl.generate_csr(name=self.ssl_blueprint.name,
                                                         common_name=self.ssl_blueprint.certificate.common_name,
                                                         sans=self.ssl_blueprint.certificate.sans,
                                                         is_ca=self.ssl_blueprint.certificate.is_ca,
                                                         key_content=key_content,
                                                         key_size=self.ssl_blueprint.certificate.private_key_size,
                                                         output_path=temp_folder.path,
                                                         **organization_parameters)

                # create tracking instance
                tracking_record_id = self.get_tracking_api().create(tracking_type=tracking.base.TrackingType.Renewal)

                # update storage api with tracking record
                self.get_storage_api().tracking_record_id = tracking_record_id

                # save all generated data
                self.save_file(
                    file_type=ssl.DataType.Blueprint,
                    file_path=self.ssl_blueprint.consolidated_blueprint_path,
                )
                self.save_file(
                    file_type=ssl.DataType.CertificateSigningRequest,
                    file_path=csr_path,
                )
                self.save_file(
                    file_type=ssl.DataType.PrivateKey,
                    file_content=key_content
                )

        return to_be_renewed

    def renew_certificate(self):
        """Perform automated renewal of the certificate using ACME protocol

        Will interact with the CA to validate ownership of the domains using ACME protocol.
        In case of any error, input TR will be automatically closed as rejected and exception logged in that TR
        In case of success, certificate is directly attached to the TR

        """
        logger.info("Processing blueprint {}".format(self.ssl_blueprint.blueprint_path))

        ca_manager = self.get_ca_manager_api()

        if not ca_manager.is_automated_renewal_supported:
            logger.info('Automated renewal not supported for {}. Nothing more to do.'.format(self.ssl_blueprint.name))
            return

        with util.TempDir() as temp_folder:
            try:
                logger.info("Start renewal process for certificate '{}'".format(self.ssl_blueprint.name))

                ########################

                # retrieve CSR from tracking or storage
                csr_path = self.get_file(file_type=ssl.DataType.CertificateSigningRequest,
                                         file_identifier='{}.{}'.format(self.ssl_blueprint.name,
                                                                        ssl.DataType.CertificateSigningRequest.value),
                                         output_folder=temp_folder.path)

                # instantiate server api for each server of the list
                # authentication for each server will be checked here
                servers_api = []
                for server_config in self.ssl_blueprint.servers:
                    servers_api.append(self.get_server_api(server_parameters=server_config))

                # request CRT to Certificate authority
                signed_crt_content = ca_manager.get_signed_certificate(
                    ssl_blueprint=self.ssl_blueprint,
                    csr_path=csr_path,
                    servers_api=servers_api,
                )

                # save certificate where requested to be used later in deployment
                self.save_file(
                    file_type=ssl.DataType.Certificate,
                    file_content=signed_crt_content
                )

                # notify that automated renewal was successful
                self.get_tracking_api().update(
                    message='Automated certificate generation successful for {}!'.format(self.ssl_blueprint.name))

            except:  # noqa
                # close record as renewal failed
                self.get_tracking_api().close_for_failure(message="Automated renewal failed: \n{}".format(
                    traceback.format_exc()))
                raise

    def deploy_certificate(self, key_path, crt_path, servers_list):
        """Deploy input SSL certificate on servers

        :param key_path: path to private key
            This is optional, if not provided, private key will be automatically retrieved from SecretServer
        :type key_path: pathlib.Path
        :param crt_path: path to certificate
        :type crt_path: pathlib.Path
        :param servers_list: list of server configuration on which to deploy the certificate.
        """

        # keep track of server with error during deployment
        deployment_error_servers = []
        # keep track of which servers have been deploy successfully
        deployment_success_servers = []

        # deploy new KEY + CRT on each server
        for server_config in servers_list:
            # instantiate correct server api based on server config
            server_api = self.get_server_api(server_parameters=server_config)
            crt_path_to_deploy = crt_path
            try:
                # we must deploy full chain of trust on this server
                # so build temp certificate that also contain CA certificate
                if server_api.deploy_full_chain:
                    certificate_content = crt_path.read_bytes()
                    # append full chain of trust to server certificate content
                    for certificate_from_chain in self.ssl_blueprint.get_chain_of_trust():
                        certificate_content += ('\n' + certificate_from_chain).encode('utf-8')
                    # create temp certificate
                    with tempfile.NamedTemporaryFile(delete=False) as tmp_certificate_file:
                        tmp_certificate_file.write(certificate_content)
                        crt_path_to_deploy = util.Path(tmp_certificate_file.name)

                # deploy key/cert on server
                server_api.deploy_cert(
                    key=key_path,
                    cert=crt_path_to_deploy,
                )
                deployment_success_servers.append(server_config)
            except exception.DeployCertificateError as ex:
                logger.error("Error when deploying certificate on {}: {}".format(server_api.get_description(), str(ex)))
                deployment_error_servers.append({'server': server_config, 'error': str(ex)})
            finally:
                # if temp certificate has just been created, clean it here
                if crt_path_to_deploy != crt_path:
                    os.remove(str(crt_path_to_deploy))

        if any(deployment_error_servers):
            self.get_tracking_api().update(
                message='New certificate successfully deployed on the following servers:\n {}\n\n'
                        'but some errors happened during deployment on following servers: {}'.format(
                            yaml.dump(data=deployment_success_servers, indent=2, default_flow_style=False),
                            yaml.dump(data=deployment_error_servers, indent=2, default_flow_style=False))
            )
        else:
            self.get_tracking_api().close_for_success(
                message='New certificate successfully deployed on all servers:\n {}'.format(
                    yaml.dump(data=servers_list, indent=2, default_flow_style=False))
            )

    def get_and_check_artifacts(
            self,
            tracking_record_id=None,
            certificate_path=None,
            private_key_path=None,
            folder=None,
    ):
        """Retrieve currently stored certificate/key and check if valid for deployment

        :param tracking_record_id: tracking record identifier
        :type tracking_record_id: str or None
        :param certificate_path: local path to SSL certificate file. Automatically retrieved if not specified.
        :type certificate_path: pathlib.Path or None
        :param private_key_path: local path to SSL certificate private key. Automatically retrieved if not specified.
        :type private_key_path: pathlib.Path or None
        :param folder: folder where artifacts will be stored.
        :type folder: pathlib.Path or None
        :return: tuple of (certificate path, private key path)
        :rtype: tuple(pathlib.Path, pathlib.Path)
        """
        # create temporary directory to save temporary files
        temp_folder = util.Path(folder or tempfile.mkdtemp())
        try:
            if tracking_record_id:
                # make sure local tracking record is up to date
                self.get_tracking_api().refresh(record_id=tracking_record_id)

            ######################################################################
            # 1. retrieve all needed artifacts
            ######################################################################
            # no certificate in input blueprint (eg: we only got global config information)
            # let's try to retrieve our blueprint from storage/tracking
            if self.ssl_blueprint.certificate is None:
                ssl_blueprint_path = self.get_file(file_type=ssl.DataType.Blueprint,
                                                   file_identifier=self.ssl_blueprint.name,
                                                   output_folder=temp_folder,
                                                   output_filename='%s.yaml' % self.ssl_blueprint.name)

                # we successfully retrieved full blueprint
                # let's refresh local ssl_blueprint from retrieved blueprint
                self.ssl_blueprint = ssl.SslBlueprint(yaml_path=ssl_blueprint_path)
                new_path = temp_folder / os.path.basename('%s.yaml' % self.ssl_blueprint.name)
                ssl_blueprint_path.rename(new_path)
                self.ssl_blueprint = ssl.SslBlueprint(yaml_path=new_path)
            logger.info('Blueprint: %s' % self.ssl_blueprint.blueprint_path.name)

            # certificate file
            certificate_full_path = temp_folder.joinpath('%s.crt' % self.ssl_blueprint.name)
            if certificate_path is not None:
                shutil.copy(str(certificate_path), str(certificate_full_path))
            if certificate_full_path.is_file():
                certificate_path = certificate_full_path
            else:
                certificate_path = self.get_file(
                    file_type=ssl.DataType.Certificate,
                    file_identifier="{}.{}".format(self.ssl_blueprint.name, ssl.DataType.Certificate.value),
                    output_folder=certificate_full_path.parent,
                    output_filename=certificate_full_path.name,
                    # if tracking_id specified, we consider tracking contains the most up to date data
                    # compare to storage so use it as primary source in that case
                    api_names=['tracking', 'storage'] if tracking_record_id else None,
                )
            logger.info('Certificate: %s' % certificate_path.name)

            # private key file
            private_key_full_path = temp_folder.joinpath('%s.key' % self.ssl_blueprint.name)
            if private_key_path is not None:
                # copy file in expected folder and only used temporary folder
                shutil.copy(str(private_key_path), str(private_key_full_path))
            if private_key_full_path.is_file():
                private_key_path = private_key_full_path
            else:
                private_key_path = self.get_file(file_type=ssl.DataType.PrivateKey,
                                                 file_identifier="{}.{}".format(
                                                     self.ssl_blueprint.name, ssl.DataType.PrivateKey.value),
                                                 output_folder=private_key_full_path.parent,
                                                 output_filename=private_key_full_path.name)
            logger.info('PrivateKey: %s' % private_key_path.name)

            ######################################################################
            # 2. perform preliminary checks on certificates before deployment
            ######################################################################
            # check that certificate matches the private key
            if not ssl.check_certificate_with_key(private_key_path, certificate_path):
                raise exception.KeyMismatch('Certificate specified does not match private key')

            # check if certificate can be used, for that use local server to reuse server api
            local_server = server.local.LocalServer(crt_name=self.ssl_blueprint.name, path=temp_folder)

            # check for expiration
            if local_server.is_expired(expiration_delay=self.ssl_blueprint.certificate.renewal_delay):
                raise exception.ExpiredCertificate(
                    "Certificate {} specified is already expired or expiring soon, renewal needed."
                    .format(self.ssl_blueprint.name))

            # check certificate match blueprint
            if not local_server.is_same(
                    common_name=self.ssl_blueprint.certificate.common_name,
                    sans=self.ssl_blueprint.certificate.sans,
                    exact_match=self.ssl_blueprint.certificate.exact_match):
                raise exception.DefinitionMismatch(
                    "Certificate {} specified does matching blueprint, renewal needed.".format(self.ssl_blueprint.name))

            # ensure chain of trust is present if full deploy requested
            full_chain_deploy_requested = any(
                    server_config.get('parameters', {}).get('deploy_full_chain', False)
                    for server_config in self.ssl_blueprint.servers
            )
            chain_of_trust = self.ssl_blueprint.get_chain_of_trust()
            if full_chain_deploy_requested and len(chain_of_trust) == 0:
                raise exception.SslBlueprintInconsistency(
                    "Deployment of full chain requested on a server but missing chain of trust.")

            # check chain of trust is valid if specified
            if len(chain_of_trust) > 0:
                ssl.check_chain_of_trust(chain_of_trust, certificate_full_path)

            return certificate_full_path, private_key_path

        # cleanup temporary files in case of any exception
        except:  # noqa: E722
            # delete folder only if not coming from user input
            if temp_folder.is_dir() and temp_folder != folder:
                shutil.rmtree(str(temp_folder), ignore_errors=True)
            raise

    def deploy(
        self,
        tracking_record_id=None,
        certificate_path=None,
        private_key_path=None,
        all_servers=False
    ):
        """Deploy certificate/key on servers

        if certificate/key file are specified in input, they will be used, else they will be retrieved from configured
        storage.

        If tracking record identifier is specified, certificate can also be retrieved from there, and this record will
        be used to track the change.
        If no tracking record specified, a new one will be created


        :param tracking_record_id: tracking record identifier
        :type tracking_record_id: str
        :param certificate_path: local path to SSL certificate file
        :type certificate_path: pathlib.Path
        :param private_key_path: local path to SSL certificate private key
        :type private_key_path: pathlib.Path
        :param all_servers: if True, deploy certificate/key on all configured servers,
            else only out of synch servers will be updated.
        :type all_servers: bool
        """
        # create temporary directory to save temporary files
        with util.TempDir() as temp_folder:
            ######################################################################
            # 1. retrieve and check all needed artifacts
            ######################################################################
            certificate_path, private_key_path = self.get_and_check_artifacts(
                tracking_record_id=tracking_record_id,
                certificate_path=certificate_path,
                private_key_path=private_key_path,
                folder=temp_folder.path,
            )

            ######################################################################
            # 2. certificate is valid, check where it should be deployed
            ######################################################################
            if all_servers is True:
                servers_to_update = self.ssl_blueprint.servers
            else:
                # check which servers are out of synch, to then deploy where it's needed without touching valid servers
                _, servers_to_update = self.get_renewal_status()

                # all servers are valid, nothing to deploy
                if len(servers_to_update) == 0:
                    logger.info("All servers up to date for certificate '%s'. Nothing to do." % self.ssl_blueprint.name)
                    return

            ######################################################################
            # 3. make sure change is properly tracked
            ######################################################################
            # if no record already specified, we don't want to modify servers without tracking
            # so create new tracking record here
            if tracking_record_id is None:
                tracking_record_id = self.get_tracking_api().create(
                    tracking_type=tracking.base.TrackingType.Synchronize,
                    servers=servers_to_update,
                )

            # update storage api with tracking record
            self.get_storage_api().tracking_record_id = tracking_record_id

            ######################################################################
            # 4. all preliminary checks are successful, start deployment on servers
            ######################################################################
            self.deploy_certificate(
                key_path=private_key_path,
                crt_path=certificate_path,
                servers_list=servers_to_update
            )

    def get_certificate_information(self, working_directory):
        """Retrieve certificate information for the blueprint.

        :param working_directory: directory in which the ssl certificate will be downloaded
        :type working_directory: pathlib.Path
        :return: SSL certificate information
        :rtype: autossl.ssl.SslCertificate
        :raise autossl.exception.NotFound: if certificate does not exist in storage
        """
        path = self.get_file(
            file_type=ssl.DataType.Certificate,
            file_identifier=self.ssl_blueprint.name+'.crt',
            output_folder=working_directory
        )
        return ssl.SslCertificate(x509_path=path)

    def get_file(self, file_type, file_identifier, output_folder, output_filename=None, api_names=None):
        """Retrieve specified stored data

        :param file_type: type of data to retrieve
        :type file_type: ssl.DataType
        :param file_identifier: identifier of the data to retrieve
        :type file_identifier: str
        :param output_folder: which folder content will be written
        :type output_folder: pathlib.Path
        :param output_filename: name of file to write (default: same than 'file_identifier' parameter)
        :type output_filename: str
        :return: local file path to the retrieved content
        :param api_names: list of api in which to search data
        :type api_names: list
        :rtype: pathlib.Path
        """
        output_file_path = output_folder / (file_identifier or output_filename)
        if output_file_path.exists():
            raise IOError("Path already exist {}".format(output_file_path))

        def __get_file_from_api(api_name_, file_type_, file_identifier_, output_file_path_):
            for file_config in self.ssl_blueprint.get_config(name=api_name_, path=['data'], default={}):
                if file_type_ == ssl.DataType(file_config['type']):
                    try:
                        file_content = self.__get_api(name=api_name_).retrieve_data(
                            name=file_identifier_,
                            data_type=file_type_,
                            **file_config.get('parameters', {})
                        )
                        if file_content is None:
                            # don't even try to write file if returned content is None, and keep searching for a match
                            continue
                        output_file_path_.write_bytes(file_content)
                        return True
                    except exception.NotFound:
                        # artifact not found in this api but it may be present in others
                        pass
            return False

        # by default use all apis
        if api_names is None:
            api_names = DEFAULT_APIS

        for api_name in api_names:
            if __get_file_from_api(api_name, file_type, file_identifier, output_file_path):
                break

        # requested data is missing
        if not output_file_path.exists():
            raise exception.NotFound("Requested file '{}' of type {} not found in tracking or storage instances."
                                     .format(file_identifier, file_type.value))

        # make sure file is named as requested in input
        if output_filename:
            output_file_path.rename(output_file_path.parent / output_filename)

        return output_file_path

    def save_file(self, file_type, file_path=None, file_content=None, api_names=None):
        """Save specified content wherever it is configured in blueprint

        :param file_type: type of data to save
        :type file_type: ssl.DataType
        :param file_path: path to a local file to save
        :type file_path: pathlib.Path
        :param file_content: content to save
        :type file_content: bytes
        :param api_names: list of api in which to save data
        :type api_names: list
        :raise IOError: if none of 'file_path' or 'file_content' parameter are specified
        """
        if not (file_path or file_content):
            raise IOError("Either one of 'file_path' or 'file_content' is expected")

        file_name = '{}.{}'.format(self.ssl_blueprint.name, file_type.value)

        # by default use all apis
        if api_names is None:
            api_names = DEFAULT_APIS

        # save all needed data
        for api_name in api_names:
            for file_config in self.ssl_blueprint.get_config(name=api_name, path=['data'], default={}):
                if file_type == ssl.DataType(file_config['type']):
                    self.__get_api(name=api_name).save_data(
                        name=file_name,
                        data_type=file_type,
                        local_path=file_path,
                        content=file_content,
                        **file_config.get('parameters', {})
                    )
                    break



"""
Tests for Base tracking implementation should not raise any error even without parent implementation
"""
import pytest

from autossl.tracking import base

from tests import util as tests_util


@pytest.fixture(scope="module")
def base_tracking():
    ssl_blueprint_name = 'tst.ov.example.com.yaml'
    ssl_blueprint_path = tests_util.DATA_PATH / ssl_blueprint_name

    base_tracking_instance = base.Tracking(ssl_blueprint_path)

    # basic checks
    assert base_tracking_instance.ssl_blueprint_path is not None

    yield base_tracking_instance


def test_create_basic(base_tracking):
    base_tracking.create(tracking_type=base.TrackingType.Renewal)


def test_save_data(base_tracking):
    base_tracking.save_data(name=None, data_type=None, local_path=None,
                            content=None, extra_param1='value1', extra_param2='value2')


def test_update(base_tracking):
    base_tracking.update(message=None)


def test_refresh(base_tracking):
    base_tracking.refresh(record_id=None)


def test_retrieve_data(base_tracking):
    base_tracking.retrieve_data(name=None, data_type=None, extra_param1='value1', extra_param2='value2')


def test_close_for_failure(base_tracking):
    base_tracking.close_for_failure(message=None)


def test_close_for_success(base_tracking):
    base_tracking.close_for_success(message=None)
